#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that validates interfaces"""
import logging
from . import bondethernet
from . import bridgedomain
from . import loopback
from . import vxlan_tunnel
from . import lcp
from . import address
from . import mac
from . import tap


def get_qinx_parent_by_name(yaml, ifname):
    """Returns the sub-interface which matches a QinAD or QinQ outer tag, or None,None
    if that sub-interface doesn't exist."""

    if not is_qinx(yaml, ifname):
        return None, None
    _qinx_ifname, qinx_iface = get_by_name(yaml, ifname)
    if not qinx_iface:
        return None, None

    qinx_encap = get_encapsulation(yaml, ifname)
    if not qinx_encap:
        return None, None

    parent_ifname, parent_iface = get_parent_by_name(yaml, ifname)
    if not parent_iface:
        return None, None

    for subid, sub_iface in parent_iface["sub-interfaces"].items():
        sub_ifname = f"{parent_ifname}.{int(subid)}"
        sub_encap = get_encapsulation(yaml, sub_ifname)
        if not sub_encap:
            continue
        if qinx_encap["dot1q"] > 0 and sub_encap["dot1q"] == qinx_encap["dot1q"]:
            return sub_ifname, sub_iface
        if qinx_encap["dot1ad"] > 0 and sub_encap["dot1ad"] == qinx_encap["dot1ad"]:
            return sub_ifname, sub_iface
    return None, None


def get_parent_by_name(yaml, ifname):
    """Returns the sub-interface's parent, or None,None if the sub-int doesn't exist."""
    if not ifname:
        return None, None

    try:
        parent_ifname, subid = ifname.split(".")
        subid = int(subid)
        iface = yaml["interfaces"][parent_ifname]
        return parent_ifname, iface
    except KeyError:
        pass
    except ValueError:
        pass
    return None, None


def get_by_lcp_name(yaml, lcpname):
    """Returns the interface or sub-interface by a given lcp name, or None,None if it does not exist"""
    if not "interfaces" in yaml:
        return None, None
    for ifname, iface in yaml["interfaces"].items():
        if "lcp" in iface and iface["lcp"] == lcpname:
            return ifname, iface
        if not "sub-interfaces" in iface:
            continue
        for subid, sub_iface in yaml["interfaces"][ifname]["sub-interfaces"].items():
            sub_ifname = f"{ifname}.{int(subid)}"
            if "lcp" in sub_iface and sub_iface["lcp"] == lcpname:
                return sub_ifname, sub_iface
    return None, None


def get_by_name(yaml, ifname):
    """Returns the interface or sub-interface by a given name, or None,None if it does not exist"""
    if "." in ifname:
        try:
            phy_ifname, subid = ifname.split(".")
            subid = int(subid)
            iface = yaml["interfaces"][phy_ifname]["sub-interfaces"][subid]
            return ifname, iface
        except ValueError:
            return None, None
        except KeyError:
            return None, None

    try:
        iface = yaml["interfaces"][ifname]
        return ifname, iface
    except KeyError:
        pass
    return None, None


def is_sub(yaml, ifname):
    """Returns True if this interface is a sub-interface"""
    _parent_ifname, parent_iface = get_parent_by_name(yaml, ifname)
    return isinstance(parent_iface, dict)


def has_sub(yaml, ifname):
    """Returns True if this interface has sub-interfaces"""
    if not "interfaces" in yaml:
        return False

    if ifname in yaml["interfaces"]:
        iface = yaml["interfaces"][ifname]
        if "sub-interfaces" in iface and len(iface["sub-interfaces"]) > 0:
            return True
    return False


def has_address(yaml, ifname):
    """Returns True if this interface or sub-interface has one or more addresses"""

    ifname, iface = get_by_name(yaml, ifname)
    if not iface:
        return False
    return "addresses" in iface


def get_l2xc_interfaces(yaml):
    """Returns a list of all interfaces that have an L2 CrossConnect"""
    ret = []
    if not "interfaces" in yaml:
        return ret
    for ifname, iface in yaml["interfaces"].items():
        if "l2xc" in iface:
            ret.append(ifname)
        if "sub-interfaces" in iface:
            for subid, sub_iface in iface["sub-interfaces"].items():
                sub_ifname = f"{ifname}.{int(subid)}"
                if "l2xc" in sub_iface:
                    ret.append(sub_ifname)

    return ret


def get_unnumbered_interfaces(yaml):
    """Returns a list of all interfaces that are unnumbered"""
    ret = []
    if not "interfaces" in yaml:
        return ret
    for ifname, iface in yaml["interfaces"].items():
        if "unnumbered" in iface:
            ret.append(ifname)
        if "sub-interfaces" in iface:
            for subid, sub_iface in iface["sub-interfaces"].items():
                sub_ifname = f"{ifname}.{int(subid)}"
                if "unnumbered" in sub_iface:
                    ret.append(sub_ifname)

    return ret


def is_l2xc_interface(yaml, ifname):
    """Returns True if this interface has an L2 CrossConnect"""

    return ifname in get_l2xc_interfaces(yaml)


def get_l2xc_target_interfaces(yaml):
    """Returns a list of all interfaces that are the target of an L2 CrossConnect"""
    ret = []
    if "interfaces" in yaml:
        for _ifname, iface in yaml["interfaces"].items():
            if "l2xc" in iface:
                ret.append(iface["l2xc"])
            if "sub-interfaces" in iface:
                for _subid, sub_iface in iface["sub-interfaces"].items():
                    if "l2xc" in sub_iface:
                        ret.append(sub_iface["l2xc"])

    return ret


def is_l2xc_target_interface(yaml, ifname):
    """Returns True if this interface is the target of an L2 CrossConnect"""

    return ifname in get_l2xc_target_interfaces(yaml)


def is_l2xc_target_interface_unique(yaml, ifname):
    """Returns True if this interface is referenced as an l2xc target zero or one times"""

    ifs = get_l2xc_target_interfaces(yaml)
    return ifs.count(ifname) < 2


def has_lcp(yaml, ifname):
    """Returns True if this interface or sub-interface has an LCP"""

    ifname, iface = get_by_name(yaml, ifname)
    if not iface:
        return False
    return "lcp" in iface


def valid_encapsulation(yaml, ifname):
    """Returns True if the sub interface has a valid encapsulation, or
    none at all"""
    ifname, iface = get_by_name(yaml, ifname)
    if not iface:
        return True
    if not "encapsulation" in iface:
        return True

    encap = iface["encapsulation"]
    if "dot1ad" in encap and "dot1q" in encap:
        return False
    if "inner-dot1q" in encap and not ("dot1ad" in encap or "dot1q" in encap):
        return False
    if "exact-match" in encap and not encap["exact-match"] and has_lcp(yaml, ifname):
        return False

    return True


def get_encapsulation(yaml, ifname):
    """Returns the encapsulation of an interface name as a fully formed dictionary:

    dot1q: int (default 0)
    dot1ad: int (default 0)
    inner-dot1q: int (default 0)
    exact-match: bool (default False)

    If the interface is not a sub-int with valid encapsulation, None is returned.
    """
    if not valid_encapsulation(yaml, ifname):
        return None

    ifname, iface = get_by_name(yaml, ifname)
    if not iface:
        return None

    _parent_ifname, parent_iface = get_parent_by_name(yaml, ifname)
    if not iface or not parent_iface:
        return None
    _parent_ifname, subid = ifname.split(".")

    dot1q = 0
    dot1ad = 0
    inner_dot1q = 0
    exact_match = False
    if not "encapsulation" in iface:
        dot1q = int(subid)
        exact_match = True
    else:
        if "dot1q" in iface["encapsulation"]:
            dot1q = iface["encapsulation"]["dot1q"]
        elif "dot1ad" in iface["encapsulation"]:
            dot1ad = iface["encapsulation"]["dot1ad"]
        if "inner-dot1q" in iface["encapsulation"]:
            inner_dot1q = iface["encapsulation"]["inner-dot1q"]
        if "exact-match" in iface["encapsulation"]:
            exact_match = iface["encapsulation"]["exact-match"]

    return {
        "dot1q": int(dot1q),
        "dot1ad": int(dot1ad),
        "inner-dot1q": int(inner_dot1q),
        "exact-match": bool(exact_match),
    }


def get_phys(yaml):
    """Return a list of all toplevel (ie. non-sub) interfaces which are
    assumed to be physical network cards, eg TenGigabitEthernet1/0/0. Note
    that derived/created interfaces such as Tunnels, BondEthernets and
    Loopbacks are not returned"""
    ret = []
    if not "interfaces" in yaml:
        return ret
    for ifname, _iface in yaml["interfaces"].items():
        if is_phy(yaml, ifname):
            ret.append(ifname)
    return ret


def is_phy(yaml, ifname):
    """Returns True if the ifname is the name of a physical network interface."""

    ifname, iface = get_by_name(yaml, ifname)
    if iface is None:
        return False
    if is_sub(yaml, ifname):
        return False

    if bondethernet.is_bondethernet(yaml, ifname):
        return False
    if loopback.is_loopback(yaml, ifname):
        return False
    if vxlan_tunnel.is_vxlan_tunnel(yaml, ifname):
        return False
    if tap.is_tap(yaml, ifname):
        return False
    return True


def get_interfaces(yaml):
    """Return a list of all interface and sub-interface names"""
    ret = []
    if not "interfaces" in yaml:
        return ret
    for ifname, iface in yaml["interfaces"].items():
        ret.append(ifname)
        if not "sub-interfaces" in iface:
            continue
        for subid, _sub_iface in iface["sub-interfaces"].items():
            ret.append(f"{ifname}.{int(subid)}")
    return ret


def get_sub_interfaces(yaml):
    """Return all interfaces which are a subinterface."""
    ret = []
    for ifname in get_interfaces(yaml):
        if is_sub(yaml, ifname):
            ret.append(ifname)
    return ret


def get_qinx_interfaces(yaml):
    """Return all interfaces which are double-tagged, either QinAD or QinQ.
    These interfaces will always have a valid encapsulation with 'inner-dot1q'
    set to non-zero.

    Note: this is always a strict subset of get_sub_interfaces()
    """
    ret = []
    for ifname in get_interfaces(yaml):
        if not is_sub(yaml, ifname):
            continue
        encap = get_encapsulation(yaml, ifname)
        if not encap:
            continue
        if encap["inner-dot1q"] > 0:
            ret.append(ifname)
    return ret


def is_qinx(yaml, ifname):
    """Returns True if the interface is a double-tagged (QinQ or QinAD) interface"""
    return ifname in get_qinx_interfaces(yaml)


def unique_encapsulation(yaml, sub_ifname):
    """Ensures that for the sub_ifname specified, there exist no other sub-ints on the
    parent with the same encapsulation."""
    new_ifname, iface = get_by_name(yaml, sub_ifname)
    parent_ifname, parent_iface = get_parent_by_name(yaml, new_ifname)
    if not iface or not parent_iface:
        return False

    sub_encap = get_encapsulation(yaml, new_ifname)
    if not sub_encap:
        return False

    ncount = 0
    for subid, _sibling_iface in parent_iface["sub-interfaces"].items():
        sibling_ifname = f"{parent_ifname}.{int(subid)}"
        sibling_encap = get_encapsulation(yaml, sibling_ifname)
        if sub_encap == sibling_encap and new_ifname != sibling_ifname:
            ncount = ncount + 1

    return ncount == 0


def is_l2(yaml, ifname):
    """Returns True if the interface is an L2XC source, L2XC target or a member of a bridgedomain"""
    if bridgedomain.is_bridge_interface(yaml, ifname):
        return True
    if is_l2xc_interface(yaml, ifname):
        return True
    if is_l2xc_target_interface(yaml, ifname):
        return True
    return False


def is_l3(yaml, ifname):
    """Returns True if the interface exists and is neither l2xc target nor bridgedomain"""
    return not is_l2(yaml, ifname)


def is_unnumbered(yaml, ifname):
    """Returns True if the interface exists and is unnumbered"""
    return ifname in get_unnumbered_interfaces(yaml)


def get_lcp(yaml, ifname):
    """Returns the LCP of the interface. If the interface is a sub-interface with L3
    enabled, synthesize it based on its parent, using smart QinQ syntax.
    Return None if no LCP can be found."""

    ifname, iface = get_by_name(yaml, ifname)
    if iface and "lcp" in iface:
        return iface["lcp"]
    return None


def get_mtu(yaml, ifname):
    """Returns MTU of the interface. If it's not set, return the parent's MTU, and
    return 1500 if no MTU was set on the sub-int or the parent."""
    ifname, iface = get_by_name(yaml, ifname)
    if iface and "mtu" in iface:
        return iface["mtu"]

    _parent_ifname, parent_iface = get_parent_by_name(yaml, ifname)
    if parent_iface and "mtu" in parent_iface:
        return parent_iface["mtu"]
    return 1500


def get_admin_state(yaml, ifname):
    """Return True if the interface admin state should be 'up'. Return False
    if it does not exist, or if it's set to 'down'."""
    ifname, iface = get_by_name(yaml, ifname)
    if not iface:
        return False
    if not "state" in iface:
        return True
    return iface["state"] == "up"


def validate_interfaces(yaml):
    """Validate the semantics of all YAML 'interfaces' entries"""
    result = True
    msgs = []
    logger = logging.getLogger("vppcfg.config")
    logger.addHandler(logging.NullHandler())

    if not "interfaces" in yaml:
        return result, msgs

    for ifname, iface in yaml["interfaces"].items():
        logger.debug(f"interface {iface}")
        if ifname.startswith("BondEthernet") and (
            None,
            None,
        ) == bondethernet.get_by_name(yaml, ifname):
            msgs.append(f"interface {ifname} does not exist in bondethernets")
            result = False
        if ifname.startswith("BondEthernet") and "mac" in iface:
            msgs.append(
                f"interface {ifname} is a member of bondethernet, cannot set MAC"
            )
            result = False
        if not "state" in iface:
            iface["state"] = "up"

        if "mac" in iface and mac.is_multicast(iface["mac"]):
            msgs.append(
                f"interface {ifname} MAC address {iface['mac']} cannot be multicast"
            )
            result = False

        if "device-type" in iface and not is_phy(yaml, ifname):
            msgs.append(f"interface {ifname} is not a PHY, cannot set device-type")
            result = False

        iface_mtu = get_mtu(yaml, ifname)
        iface_lcp = get_lcp(yaml, ifname)
        iface_address = has_address(yaml, ifname)

        if ifname.startswith("tap"):
            _tap_ifname, tap_iface = tap.get_by_name(yaml, ifname)
            if not tap_iface:
                msgs.append(f"interface {ifname} is a TAP but does not exist in taps")
                result = False
            elif "mtu" in tap_iface["host"]:
                host_mtu = tap_iface["host"]["mtu"]
                if host_mtu != iface_mtu:
                    msgs.append(
                        f"interface {ifname} is a TAP so its MTU {int(iface_mtu)} must match host MTU {int(host_mtu)}"
                    )
                    result = False
            if iface_address:
                msgs.append(f"interface {ifname} is a TAP so it cannot have an address")
                result = False
            if iface_lcp:
                msgs.append(f"interface {ifname} is a TAP so it cannot have an LCP")
                result = False
            if has_sub(yaml, ifname):
                msgs.append(
                    f"interface {ifname} is a TAP so it cannot have sub-interfaces"
                )
                result = False

        if is_l2(yaml, ifname) and iface_lcp:
            msgs.append(
                f"interface {ifname} is in L2 mode but has LCP name {iface_lcp}"
            )
            result = False
        if is_l2(yaml, ifname) and iface_address:
            msgs.append(f"interface {ifname} is in L2 mode but has an address")
            result = False
        if iface_lcp and not lcp.is_unique(yaml, iface_lcp):
            msgs.append(
                f"interface {ifname} does not have a unique LCP name {iface_lcp}"
            )
            result = False

        if "unnumbered" in iface:
            target = iface["unnumbered"]
            _, target_iface = loopback.get_by_name(yaml, target)
            if not target_iface:
                _, target_iface = get_by_name(yaml, target)
            if not target_iface:
                msgs.append(
                    f"interface {ifname} unnumbered target {target} does not exist"
                )
                result = False
            if is_l2(yaml, target):
                msgs.append(
                    f"interface {ifname} unnumbered target {target} cannot be in L2 mode"
                )
                result = False
            if is_unnumbered(yaml, target) or loopback.is_unnumbered(yaml, target):
                msgs.append(
                    f"interface {ifname} unnumbered target {target} cannot also be unnumbered"
                )
                result = False
            if ifname == target:
                msgs.append(
                    f"interface {ifname} unnumbered target cannot point to itself"
                )
                result = False
            if has_address(yaml, ifname):
                msgs.append(
                    f"interface {ifname} cannot also have addresses when it is unnumbered"
                )
                result = False

        if "addresses" in iface:
            for addr in iface["addresses"]:
                if not address.is_allowed(yaml, ifname, iface["addresses"], addr):
                    msgs.append(
                        f"interface {ifname} IP address {addr} conflicts with another"
                    )
                    result = False
                if not address.is_canonical(addr):
                    canonical = address.get_canonical(addr)
                    msgs.append(
                        f"interface {ifname} IP address {addr} is not canonical, use {canonical}"
                    )
                    result = False

        if "l2xc" in iface:
            if has_sub(yaml, ifname):
                msgs.append(
                    f"interface {ifname} has l2xc so it cannot have sub-interfaces"
                )
                result = False
            if iface_lcp:
                msgs.append(f"interface {ifname} has l2xc so it cannot have an LCP")
                result = False
            if iface_address:
                msgs.append(f"interface {ifname} has l2xc so it cannot have an address")
                result = False
            if (None, None) == get_by_name(yaml, iface["l2xc"]):
                msgs.append(
                    f"interface {ifname} l2xc target {iface['l2xc']} does not exist"
                )
                result = False
            if iface["l2xc"] == ifname:
                msgs.append(f"interface {ifname} l2xc target cannot be itself")
                result = False
            target_mtu = get_mtu(yaml, iface["l2xc"])
            if target_mtu != iface_mtu:
                msgs.append(
                    f"interface {ifname} l2xc target MTU {int(target_mtu)} does not match source MTU {int(iface_mtu)}"
                )
                result = False
            if not is_l2xc_target_interface_unique(yaml, iface["l2xc"]):
                msgs.append(
                    f"interface {ifname} l2xc target {iface['l2xc']} is not unique"
                )
                result = False
            if bridgedomain.is_bridge_interface(yaml, iface["l2xc"]):
                msgs.append(
                    f"interface {ifname} l2xc target {iface['l2xc']} is in a bridgedomain"
                )
                result = False
            if has_lcp(yaml, iface["l2xc"]):
                msgs.append(
                    f"interface {ifname} l2xc target {iface['l2xc']} cannot have an LCP"
                )
                result = False
            if has_address(yaml, iface["l2xc"]):
                msgs.append(
                    f"interface {ifname} l2xc target {iface['l2xc']} cannot have an address"
                )
                result = False

        if has_sub(yaml, ifname):
            for sub_id, sub_iface in yaml["interfaces"][ifname][
                "sub-interfaces"
            ].items():
                logger.debug(f"sub-interface {sub_iface}")
                sub_ifname = f"{ifname}.{int(sub_id)}"
                if not sub_iface:
                    msgs.append(f"sub-interface {sub_ifname} has no config")
                    result = False
                    continue

                if not "state" in sub_iface:
                    sub_iface["state"] = "up"
                if sub_iface["state"] == "up" and iface["state"] == "down":
                    msgs.append(
                        f"sub-interface {sub_ifname} cannot be up if parent {ifname} is down"
                    )
                    result = False

                sub_mtu = get_mtu(yaml, sub_ifname)
                if sub_mtu > iface_mtu:
                    msgs.append(
                        f"sub-interface {sub_ifname} has MTU {int(sub_iface['mtu'])} higher than parent {ifname} MTU {int(iface_mtu)}"
                    )
                    result = False
                if is_qinx(yaml, sub_ifname):
                    mid_ifname, mid_iface = get_qinx_parent_by_name(yaml, sub_ifname)
                    mid_mtu = get_mtu(yaml, mid_ifname)
                    if sub_mtu > mid_mtu:
                        msgs.append(
                            f"sub-interface {sub_ifname} has MTU {int(sub_iface['mtu'])} higher than parent {mid_ifname} MTU {int(mid_mtu)}"
                        )
                        result = False

                sub_lcp = get_lcp(yaml, sub_ifname)
                if is_l2(yaml, sub_ifname) and sub_lcp:
                    msgs.append(
                        f"sub-interface {sub_ifname} is in L2 mode but has LCP name {sub_lcp}"
                    )
                    result = False
                if sub_lcp and not lcp.is_unique(yaml, sub_lcp):
                    msgs.append(
                        f"sub-interface {sub_ifname} does not have a unique LCP name {sub_lcp}"
                    )
                    result = False
                if sub_lcp and not iface_lcp:
                    msgs.append(
                        f"sub-interface {sub_ifname} has LCP name {sub_lcp} but {ifname} does not have an LCP"
                    )
                    result = False
                if sub_lcp and is_qinx(yaml, sub_ifname):
                    mid_ifname, mid_iface = get_qinx_parent_by_name(yaml, sub_ifname)
                    if not mid_iface:
                        msgs.append(
                            f"sub-interface {sub_ifname} is QinX and has LCP name {sub_lcp} which requires a parent"
                        )
                        result = False
                    elif not get_lcp(yaml, mid_ifname):
                        msgs.append(
                            f"sub-interface {sub_ifname} is QinX and has LCP name {sub_lcp} but {mid_ifname} does not have an LCP"
                        )
                        result = False

                encap = get_encapsulation(yaml, sub_ifname)
                if sub_lcp and (not encap or not encap["exact-match"]):
                    msgs.append(
                        f"sub-interface {sub_ifname} has LCP name {sub_lcp} but its encapsulation is not exact-match"
                    )
                    result = False

                if "unnumbered" in sub_iface:
                    target = sub_iface["unnumbered"]
                    _, target_iface = loopback.get_by_name(yaml, target)
                    if not target_iface:
                        _, target_iface = get_by_name(yaml, target)
                    if not target_iface:
                        msgs.append(
                            f"sub-interface {sub_ifname} unnumbered target {target} does not exist"
                        )
                        result = False
                    if is_l2(yaml, target):
                        msgs.append(
                            f"sub-interface {sub_ifname} unnumbered target {target} cannot be in L2 mode"
                        )
                        result = False
                    if is_unnumbered(yaml, target) or loopback.is_unnumbered(
                        yaml, target
                    ):
                        msgs.append(
                            f"sub-interface {sub_ifname} unnumbered target {target} cannot also be unnumbered"
                        )
                        result = False
                    if sub_ifname == target:
                        msgs.append(
                            f"sub-interface {sub_ifname} unnumbered target cannot point to itself"
                        )
                        result = False
                    if has_address(yaml, sub_ifname):
                        msgs.append(
                            f"sub-interface {sub_ifname} cannot also have addresses when it is unnumbered"
                        )
                        result = False

                if has_address(yaml, sub_ifname):
                    if not encap or not encap["exact-match"]:
                        msgs.append(
                            f"sub-interface {sub_ifname} has an address but its encapsulation is not exact-match"
                        )
                        result = False
                    if is_l2(yaml, sub_ifname):
                        msgs.append(
                            f"sub-interface {sub_ifname} is in L2 mode but has an address"
                        )
                        result = False
                    for addr in sub_iface["addresses"]:
                        if not address.is_allowed(
                            yaml, sub_ifname, sub_iface["addresses"], addr
                        ):
                            msgs.append(
                                f"sub-interface {sub_ifname} IP address {addr} conflicts with another"
                            )
                            result = False
                if not valid_encapsulation(yaml, sub_ifname):
                    msgs.append(f"sub-interface {sub_ifname} has invalid encapsulation")
                    result = False
                elif not unique_encapsulation(yaml, sub_ifname):
                    msgs.append(
                        f"sub-interface {sub_ifname} does not have unique encapsulation"
                    )
                    result = False
                if "l2xc" in sub_iface:
                    if has_lcp(yaml, sub_ifname):
                        msgs.append(
                            f"sub-interface {sub_ifname} has l2xc so it cannot have an LCP"
                        )
                        result = False
                    if has_address(yaml, sub_ifname):
                        msgs.append(
                            f"sub-interface {sub_ifname} has l2xc so it cannot have an address"
                        )
                        result = False
                    if (None, None) == get_by_name(yaml, sub_iface["l2xc"]):
                        msgs.append(
                            f"sub-interface {sub_ifname} l2xc target {sub_iface['l2xc']} does not exist"
                        )
                        result = False
                    if sub_iface["l2xc"] == sub_ifname:
                        msgs.append(
                            f"sub-interface {sub_ifname} l2xc target cannot be itself"
                        )
                        result = False
                    target_mtu = get_mtu(yaml, sub_iface["l2xc"])
                    if target_mtu != sub_mtu:
                        msgs.append(
                            f"sub-interface {sub_ifname} l2xc target MTU {int(target_mtu)} does not match source MTU {int(sub_mtu)}"
                        )
                        result = False
                    if not is_l2xc_target_interface_unique(yaml, sub_iface["l2xc"]):
                        msgs.append(
                            f"sub-interface {sub_ifname} l2xc target {sub_iface['l2xc']} is not unique"
                        )
                        result = False
                    if bridgedomain.is_bridge_interface(yaml, sub_iface["l2xc"]):
                        msgs.append(
                            f"sub-interface {sub_ifname} l2xc target {sub_iface['l2xc']} is in a bridgedomain"
                        )
                        result = False
                    if has_lcp(yaml, sub_iface["l2xc"]):
                        msgs.append(
                            f"sub-interface {sub_ifname} l2xc target {sub_iface['l2xc']} cannot have an LCP"
                        )
                        result = False
                    if has_address(yaml, sub_iface["l2xc"]):
                        msgs.append(
                            f"sub-interface {sub_ifname} l2xc target {sub_iface['l2xc']} cannot have an address"
                        )
                        result = False

    return result, msgs


def is_mpls(yaml, ifname):
    """Returns True if the interface exists and has mpls enabled. Returns false otherwise."""
    _, iface = get_by_name(yaml, ifname)
    if iface and "mpls" in iface and iface["mpls"]:
        return True
    return False



# Copyright (c) 2025 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""
Planner create operations - handles creation of objects in VPP that exist in config but not in VPP.
"""

from vppcfg.config import loopback
from vppcfg.config import interface
from vppcfg.config import bondethernet
from vppcfg.config import bridgedomain
from vppcfg.config import vxlan_tunnel
from vppcfg.config import tap
from .planner_base import PlannerBase


class PlannerCreateOperations(PlannerBase):  # pylint: disable=too-few-public-methods
    """Mixin class providing create operations for the Planner."""

    def create(self):
        """Create all objects in VPP that occur in the config but not in VPP. For an indepth
        explanation of how and why this particular creation order is chosen, see README.md
        section on Planning."""
        ret = True
        if not self._create_loopbacks():
            self.logger.warning("Could not create Loopbacks in VPP")
            ret = False
        if not self._create_bondethernets():
            self.logger.warning("Could not create BondEthernets in VPP")
            ret = False
        if not self._create_vxlan_tunnels():
            self.logger.warning("Could not create VXLAN Tunnels in VPP")
            ret = False
        if not self._create_taps():
            self.logger.warning("Could not create TAPs in VPP")
            ret = False
        if not self._create_sub_interfaces():
            self.logger.warning("Could not create Sub Interfaces in VPP")
            ret = False
        if not self._create_bridgedomains():
            self.logger.warning("Could not create BridgeDomains in VPP")
            ret = False
        if not self._create_lcps():
            self.logger.warning("Could not create LCPs in VPP")
            ret = False
        return ret

    def _create_loopbacks(self):
        """Create all loopbacks that occur in the config but not in VPP"""
        for ifname in loopback.get_loopbacks(self.cfg):
            if ifname in self.vpp.cache["interface_names"]:
                continue
            instance = int(ifname[4:])
            cli = f"create loopback interface instance {int(instance)}"
            ifname, iface = loopback.get_by_name(self.cfg, ifname)
            if "mac" in iface:
                cli += f" mac {iface['mac']}"
            self.cli["create"].append(cli)
        return True

    def _create_bondethernets(self):
        """Create all bondethernets that occur in the config but not in VPP"""
        for ifname in bondethernet.get_bondethernets(self.cfg):
            if ifname in self.vpp.cache["interface_names"]:
                continue
            ifname, iface = bondethernet.get_by_name(self.cfg, ifname)
            instance = int(ifname[12:])
            mode = bondethernet.get_mode(self.cfg, ifname)
            cli = f"create bond id {int(instance)} mode {mode}"
            loadbalance = bondethernet.get_lb(self.cfg, ifname)
            if loadbalance:
                cli += f" load-balance {loadbalance}"
            if "mac" in iface:
                cli += f" hw-addr {iface['mac']}"
            self.cli["create"].append(cli)
        return True

    def _create_vxlan_tunnels(self):
        """Create all vxlan_tunnels that occur in the config but not in VPP"""
        for ifname in vxlan_tunnel.get_vxlan_tunnels(self.cfg):
            if ifname in self.vpp.cache["interface_names"]:
                continue
            ifname, iface = vxlan_tunnel.get_by_name(self.cfg, ifname)
            instance = int(ifname[12:])
            cli = (
                f"create vxlan tunnel src {iface['local']} dst {iface['remote']} "
                f"instance {instance} vni {iface['vni']} decap-next l2"
            )
            self.cli["create"].append(cli)
        return True

    def _create_sub_interfaces(self):
        """Create all sub-interfaces that occur in the config but not in VPP"""
        ## First create 1-tag (Dot1Q/Dot1AD), and then create 2-tag (Qin*) sub-interfaces
        for do_qinx in [False, True]:
            for ifname in interface.get_sub_interfaces(self.cfg):
                if not do_qinx == interface.is_qinx(self.cfg, ifname):
                    continue

                ifname, _iface = interface.get_by_name(self.cfg, ifname)
                if ifname in self.vpp.cache["interface_names"]:
                    continue

                ## Assemble the encapsulation string
                encap = interface.get_encapsulation(self.cfg, ifname)
                if encap["dot1ad"] > 0:
                    encapstr = f"dot1ad {int(encap['dot1ad'])}"
                else:
                    encapstr = f"dot1q {int(encap['dot1q'])}"
                if do_qinx:
                    encapstr += f" inner-dot1q {int(encap['inner-dot1q'])}"
                if encap["exact-match"]:
                    encapstr += " exact-match"
                parent, subid = ifname.split(".")
                cli = f"create sub {parent} {int(int(subid))} {encapstr}"
                self.cli["create"].append(cli)
        return True

    def _create_taps(self):
        """Create all taps that occur in the config but not in VPP"""
        for ifname in tap.get_taps(self.cfg):
            ifname, iface = tap.get_by_name(self.cfg, ifname)
            if ifname in self.vpp.cache["interface_names"]:
                continue
            instance = int(ifname[3:])
            cli = f"create tap id {int(instance)} host-if-name {iface['host']['name']}"
            if "mac" in iface["host"]:
                cli += f" host-mac-addr {iface['host']['mac']}"
            if "namespace" in iface["host"]:
                cli += f" host-ns {int(iface['host']['namespace'])}"
            if "bridge" in iface["host"]:
                cli += f" host-bridge {iface['host']['bridge']}"
            if "mtu" in iface["host"]:
                cli += f" host-mtu-size {int(iface['host']['mtu'])}"
            if "rx-ring-size" in iface:
                cli += f" rx-ring-size {int(iface['rx-ring-size'])}"
            if "tx-ring-size" in iface:
                cli += f" tx-ring-size {int(iface['tx-ring-size'])}"
            self.cli["create"].append(cli)

        return True

    def _create_bridgedomains(self):
        """Create all bridgedomains that occur in the config but not in VPP"""
        for ifname in bridgedomain.get_bridgedomains(self.cfg):
            ifname, _iface = bridgedomain.get_by_name(self.cfg, ifname)
            instance = int(ifname[2:])
            settings = bridgedomain.get_settings(self.cfg, ifname)
            if instance in self.vpp.cache["bridgedomains"]:
                continue
            cli = f"create bridge-domain {instance}"
            if not settings["learn"]:
                cli += " learn 0"
            if not settings["unicast-flood"]:
                cli += " flood 0"
            if not settings["unknown-unicast-flood"]:
                cli += " uu-flood 0"
            if not settings["unicast-forward"]:
                cli += " forward 0"
            if settings["arp-termination"]:
                cli += " arp-term 1"
            if settings["arp-unicast-forward"]:
                cli += " arp-ufwd 1"
            if settings["mac-age-minutes"] > 0:
                cli += f" mac-age {int(settings['mac-age-minutes'])}"
            self.cli["create"].append(cli)
        return True

    def _create_lcps(self):
        """Create all LCPs that occur in the config but not in VPP"""
        lcpnames = [
            self.vpp.cache["lcps"][x].host_if_name for x in self.vpp.cache["lcps"]
        ]

        ## First create untagged ...
        for ifname in loopback.get_loopbacks(self.cfg) + interface.get_interfaces(
            self.cfg
        ):
            if interface.is_sub(self.cfg, ifname):
                continue

            if ifname.startswith("loop"):
                ifname, iface = loopback.get_by_name(self.cfg, ifname)
            else:
                ifname, iface = interface.get_by_name(self.cfg, ifname)
            if not "lcp" in iface:
                continue
            if iface["lcp"] in lcpnames:
                continue
            cli = f"lcp create {ifname} host-if {iface['lcp']}"
            self.cli["create"].append(cli)

        ## ... then 1-tag (Dot1Q/Dot1AD), and then create 2-tag (Qin*) LCPs
        for do_qinx in [False, True]:
            for ifname in interface.get_sub_interfaces(self.cfg):
                if not do_qinx == interface.is_qinx(self.cfg, ifname):
                    continue
                ifname, iface = interface.get_by_name(self.cfg, ifname)
                if not "lcp" in iface:
                    continue
                if iface["lcp"] in lcpnames:
                    continue
                cli = f"lcp create {ifname} host-if {iface['lcp']}"
                self.cli["create"].append(cli)
        return True



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""Unit tests for addresses"""
import unittest
import yaml
from . import address


class TestAddressMethods(unittest.TestCase):
    def test_get_canonical(self):
        self.assertEqual(address.get_canonical("0.0.0.0"), "0.0.0.0")
        self.assertEqual(address.get_canonical("0.0.0.0/0"), "0.0.0.0/0")
        self.assertEqual(address.get_canonical("192.168.1.1"), "192.168.1.1")
        self.assertEqual(address.get_canonical("192.168.1.1/32"), "192.168.1.1/32")
        self.assertEqual(address.get_canonical("2001:db8::1"), "2001:db8::1")
        self.assertEqual(address.get_canonical("2001:db8::1/64"), "2001:db8::1/64")

        self.assertEqual(address.get_canonical("2001:dB8::1/128"), "2001:db8::1/128")
        self.assertEqual(address.get_canonical("2001:db8:0::1/128"), "2001:db8::1/128")
        self.assertEqual(address.get_canonical("2001:db8::0:1"), "2001:db8::1")

    def test_is_canonical(self):
        self.assertTrue(address.is_canonical("0.0.0.0"))
        self.assertTrue(address.is_canonical("0.0.0.0/0"))
        self.assertTrue(address.is_canonical("192.168.1.1"))
        self.assertTrue(address.is_canonical("192.168.1.1/32"))
        self.assertTrue(address.is_canonical("2001:db8::1"))
        self.assertTrue(address.is_canonical("2001:db8::1/64"))

        self.assertFalse(address.is_canonical("2001:dB8::1/128"))  # Capitals
        self.assertFalse(address.is_canonical("2001:db8:0::1/128"))  # Spurious 0
        self.assertFalse(address.is_canonical("2001:db8::0:1"))  # Spurious 0



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""Unit tests for bridgedomains"""
import unittest
import yaml
from . import bridgedomain
from .unittestyaml import UnitTestYaml


class TestBridgeDomainMethods(unittest.TestCase):
    def setUp(self):
        with UnitTestYaml("test_bridgedomain.yaml") as f:
            self.cfg = yaml.load(f, Loader=yaml.FullLoader)

    def test_get_by_name(self):
        ifname, iface = bridgedomain.get_by_name(self.cfg, "bd10")
        self.assertIsNotNone(iface)
        self.assertEqual("bd10", ifname)
        self.assertEqual(iface["mtu"], 3000)
        self.assertIn("BondEthernet0", iface["interfaces"])

        ifname, iface = bridgedomain.get_by_name(self.cfg, "bd-notexist")
        self.assertIsNone(iface)
        self.assertIsNone(ifname)

    def test_is_bridgedomain(self):
        self.assertTrue(bridgedomain.is_bridgedomain(self.cfg, "bd10"))
        self.assertTrue(bridgedomain.is_bridgedomain(self.cfg, "bd11"))
        self.assertFalse(bridgedomain.is_bridgedomain(self.cfg, "bd-notexist"))
        self.assertFalse(bridgedomain.is_bridgedomain(self.cfg, "GigabitEthernet1/0/0"))

    def test_members(self):
        self.assertTrue(
            bridgedomain.is_bridge_interface(self.cfg, "GigabitEthernet1/0/0")
        )
        self.assertTrue(
            bridgedomain.is_bridge_interface(self.cfg, "GigabitEthernet2/0/0.100")
        )
        self.assertFalse(
            bridgedomain.is_bridge_interface(self.cfg, "GigabitEthernet3/0/0")
        )
        self.assertFalse(
            bridgedomain.is_bridge_interface(self.cfg, "GigabitEthernet3/0/0.100")
        )

    def test_unique(self):
        self.assertFalse(
            bridgedomain.is_bridge_interface_unique(self.cfg, "GigabitEthernet1/0/0")
        )
        self.assertTrue(
            bridgedomain.is_bridge_interface_unique(
                self.cfg, "GigabitEthernet2/0/0.100"
            )
        )

    def test_enumerators(self):
        ifs = bridgedomain.get_bridge_interfaces(self.cfg)
        self.assertEqual(len(ifs), 8)
        self.assertIn("BondEthernet0", ifs)
        self.assertIn("GigabitEthernet1/0/0", ifs)
        self.assertIn("GigabitEthernet2/0/0.100", ifs)

    def test_bvi_unique(self):
        self.assertTrue(bridgedomain.bvi_unique(self.cfg, "loop0"))
        self.assertFalse(bridgedomain.bvi_unique(self.cfg, "loop1"))
        self.assertTrue(bridgedomain.bvi_unique(self.cfg, "loop2"))

    def test_get_bridgedomains(self):
        ifs = bridgedomain.get_bridgedomains(self.cfg)
        self.assertEqual(len(ifs), 6)

    def test_get_settings(self):
        settings = bridgedomain.get_settings(self.cfg, "bd1")
        self.assertIsNone(settings)

        settings = bridgedomain.get_settings(self.cfg, "bd10")
        self.assertTrue(settings["learn"])
        self.assertTrue(settings["unknown-unicast-flood"])
        self.assertTrue(settings["unicast-flood"])
        self.assertEqual(settings["mac-age-minutes"], 0)

        settings = bridgedomain.get_settings(self.cfg, "bd11")
        self.assertTrue(settings["learn"])
        self.assertFalse(settings["unknown-unicast-flood"])
        self.assertFalse(settings["unicast-flood"])
        self.assertEqual(settings["mac-age-minutes"], 10)



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that validates loopbacks"""
import logging
from . import lcp
from . import address
from . import mac
from . import interface


def get_loopbacks(yaml):
    """Return a list of all loopbacks."""
    ret = []
    if "loopbacks" in yaml:
        for ifname, _iface in yaml["loopbacks"].items():
            ret.append(ifname)
    return ret


def get_by_lcp_name(yaml, lcpname):
    """Returns the loopback by a given lcp name, or None,None if it does not exist"""
    if not "loopbacks" in yaml:
        return None, None
    for ifname, iface in yaml["loopbacks"].items():
        if "lcp" in iface and iface["lcp"] == lcpname:
            return ifname, iface
    return None, None


def get_by_name(yaml, ifname):
    """Return the loopback by name, if it exists. Return None otherwise."""
    try:
        if ifname in yaml["loopbacks"]:
            return ifname, yaml["loopbacks"][ifname]
    except KeyError:
        pass
    return None, None


def is_loopback(yaml, ifname):
    """Returns True if the interface name is an existing loopback."""
    ifname, iface = get_by_name(yaml, ifname)
    return iface is not None


def get_unnumbered_loopbacks(yaml):
    """Returns a list of all loopbacks that are unnumbered"""
    ret = []
    if not "loopbacks" in yaml:
        return ret
    for ifname, iface in yaml["loopbacks"].items():
        if "unnumbered" in iface:
            ret.append(ifname)

    return ret


def is_unnumbered(yaml, ifname):
    """Returns True if the loopback exists and is unnumbered"""
    return ifname in get_unnumbered_loopbacks(yaml)


def has_address(yaml, ifname):
    """Returns True if this loopback has one or more addresses"""

    ifname, iface = get_by_name(yaml, ifname)
    if not iface:
        return False
    return "addresses" in iface


def validate_loopbacks(yaml):
    """Validate the semantics of all YAML 'loopbacks' entries"""
    result = True
    msgs = []
    logger = logging.getLogger("vppcfg.config")
    logger.addHandler(logging.NullHandler())

    if not "loopbacks" in yaml:
        return result, msgs

    for ifname, iface in yaml["loopbacks"].items():
        logger.debug(f"loopback {iface}")
        instance = int(ifname[4:])
        if instance > 4095:
            msgs.append(
                f"loopback {ifname} has instance {int(instance)} which is too large"
            )
            result = False
        if "lcp" in iface and not lcp.is_unique(yaml, iface["lcp"]):
            msgs.append(
                f"loopback {ifname} does not have a unique LCP name {iface['lcp']}"
            )
            result = False
        if "unnumbered" in iface:
            target = iface["unnumbered"]
            _, target_iface = get_by_name(yaml, target)
            if not target_iface:
                _, target_iface = interface.get_by_name(yaml, target)
            if not target_iface:
                msgs.append(
                    f"loopback {ifname} unnumbered target {target} does not exist"
                )
                result = False
            if is_unnumbered(yaml, target) or interface.is_unnumbered(yaml, target):
                msgs.append(
                    f"loopback {ifname} unnumbered target {target} cannot also be unnumbered"
                )
                result = False
            if ifname == target:
                msgs.append(
                    f"loopback {ifname} unnumbered target cannot point to itself"
                )
                result = False
            if has_address(yaml, ifname):
                msgs.append(
                    f"loopback {ifname} cannot also have addresses when it is unnumbered"
                )
                result = False
        if "addresses" in iface:
            for addr in iface["addresses"]:
                if not address.is_allowed(yaml, ifname, iface["addresses"], addr):
                    msgs.append(
                        f"loopback {ifname} IP address {addr} conflicts with another"
                    )
                    result = False
                if not address.is_canonical(addr):
                    canonical = address.get_canonical(addr)
                    msgs.append(
                        f"loopback {ifname} IP address {addr} is not canonical, use {canonical}"
                    )
                    result = False
        if "mac" in iface and mac.is_multicast(iface["mac"]):
            msgs.append(
                f"loopback {ifname} MAC address {iface['mac']} cannot be multicast"
            )
            result = False

    return result, msgs


def is_mpls(yaml, ifname):
    """Returns True if the loopback exists and has mpls enabled. Returns false otherwise."""
    _, iface = get_by_name(yaml, ifname)
    if iface and "mpls" in iface and iface["mpls"]:
        return True
    return False



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that handles bondethernets"""
import logging
from . import interface
from . import mac


def get_bondethernets(yaml):
    """Return a list of all bondethernets."""
    ret = []
    if "bondethernets" in yaml:
        for ifname, _iface in yaml["bondethernets"].items():
            ret.append(ifname)
    return ret


def get_by_name(yaml, ifname):
    """Return the BondEthernet by name, if it exists. Return None,None otherwise."""
    try:
        if ifname in yaml["bondethernets"]:
            return ifname, yaml["bondethernets"][ifname]
    except KeyError:
        pass
    return None, None


def is_bondethernet(yaml, ifname):
    """Returns True if the interface name is an existing BondEthernet."""
    ifname, iface = get_by_name(yaml, ifname)
    return iface is not None


def is_bond_member(yaml, ifname):
    """Returns True if this interface is a member of a BondEthernet."""
    if not "bondethernets" in yaml:
        return False

    for _bond, iface in yaml["bondethernets"].items():
        if not "interfaces" in iface:
            continue
        if ifname in iface["interfaces"]:
            return True
    return False


def get_mode(yaml, ifname):
    """Return the mode of the BondEthernet as a string, defaulting to 'lacp'
    if no mode is given. Return None if the bond interface doesn't exist.

    Return values: 'round-robin','active-backup','broadcast','lacp','xor'
    """
    ifname, iface = get_by_name(yaml, ifname)
    if not iface:
        return None

    if not "mode" in iface:
        return "lacp"
    return iface["mode"]


def mode_to_int(mode):
    """Returns the integer representation in VPP of a given bondethernet mode,
    or -1 if 'mode' is not a valid string.

    See src/vnet/bonding/bond.api and schema.yaml for valid pairs."""

    ret = {"round-robin": 1, "active-backup": 2, "xor": 3, "broadcast": 4, "lacp": 5}
    try:
        return ret[mode]
    except KeyError:
        pass
    return -1


def int_to_mode(mode):
    """Returns the string representation in VPP of a given bondethernet mode,
    or "" if 'mode' is not a valid id.

    See src/vnet/bonding/bond.api and schema.yaml for valid pairs."""

    ret = {1: "round-robin", 2: "active-backup", 3: "xor", 4: "broadcast", 5: "lacp"}
    try:
        return ret[mode]
    except KeyError:
        pass
    return ""


def get_lb(yaml, ifname):
    """Return the loadbalance strategy of the BondEthernet as a string. Only
    'xor' and 'lacp' modes have loadbalance strategies, so return None if
    those modes are not used.

    Return values: 'l2', 'l23', 'l34', with 'l34' being the default if
    the bond is in xor/lacp mode without a load-balance strategy set
    explicitly."""
    ifname, iface = get_by_name(yaml, ifname)
    if not iface:
        return None
    mode = get_mode(yaml, ifname)
    if mode not in ["xor", "lacp"]:
        return None

    if not "load-balance" in iface:
        return "l34"
    return iface["load-balance"]


def lb_to_int(loadbalance):
    """Returns the integer representation in VPP of a given load-balance strategy,
    or -1 if 'lb' is not a valid string.

    See src/vnet/bonding/bond.api and schema.yaml for valid pairs, although
    bond.api defined more than we use in vppcfg."""

    ret = {
        "l2": 0,
        "l34": 1,
        "l23": 2,
        "round-robin": 3,
        "broadcast": 4,
        "active-backup": 5,
    }
    try:
        return ret[loadbalance]
    except KeyError:
        pass
    return -1


def int_to_lb(loadbalance):
    """Returns the string representation in VPP of a given load-balance strategy,
    or "" if 'lb' is not a valid int.

    See src/vnet/bonding/bond.api and schema.yaml for valid pairs, although
    bond.api defined more than we use in vppcfg."""

    ret = {
        0: "l2",
        1: "l34",
        2: "l23",
        3: "round-robin",
        4: "broadcast",
        5: "active-backup",
    }
    try:
        return ret[loadbalance]
    except KeyError:
        pass
    return ""


def validate_bondethernets(yaml):
    """Validate the semantics of all YAML 'bondethernets' entries"""
    result = True
    msgs = []
    logger = logging.getLogger("vppcfg.config")
    logger.addHandler(logging.NullHandler())

    if not "bondethernets" in yaml:
        return result, msgs

    for ifname, iface in yaml["bondethernets"].items():
        logger.debug(f"bondethernet {ifname}: {iface}")
        bond_ifname, bond_iface = interface.get_by_name(yaml, ifname)
        bond_mtu = 1500
        if not bond_iface:
            msgs.append(f"bondethernet {ifname} does not exist in interfaces")
            result = False
        else:
            bond_mtu = interface.get_mtu(yaml, bond_ifname)
        instance = int(ifname[12:])
        if instance > 4294967294:
            msgs.append(
                f"bondethernet {ifname} has instance {int(instance)} which is too large"
            )
            result = False
        if (
            get_mode(yaml, bond_ifname) not in ["xor", "lacp"]
            and "load-balance" in iface
        ):
            msgs.append(
                f"bondethernet {ifname} can only have load-balance if in mode XOR or LACP"
            )
            result = False
        if "mac" in iface and mac.is_multicast(iface["mac"]):
            msgs.append(
                f"bondethernet {ifname} MAC address {iface['mac']} cannot be multicast"
            )
            result = False

        if not "interfaces" in iface:
            continue

        for member in iface["interfaces"]:
            if (None, None) == interface.get_by_name(yaml, member):
                msgs.append(f"bondethernet {ifname} member {member} does not exist")
                result = False
                continue

            if interface.has_sub(yaml, member):
                msgs.append(
                    f"bondethernet {ifname} member {member} has sub-interface(s)"
                )
                result = False
            if interface.has_lcp(yaml, member):
                msgs.append(f"bondethernet {ifname} member {member} has an LCP")
                result = False
            if interface.has_address(yaml, member):
                msgs.append(f"bondethernet {ifname} member {member} has an address")
                result = False
            member_mtu = interface.get_mtu(yaml, member)
            if member_mtu != bond_mtu:
                msgs.append(
                    f"bondethernet {ifname} member {member} MTU {int(member_mtu)} does not match BondEthernet MTU {int(bond_mtu)}"
                )
                result = False
    return result, msgs



"""vppcfg setuptools setup.py for pip and deb pkg installations"""

import os
from setuptools import setup

# Read version from _version.py
version_file = os.path.join(os.path.dirname(__file__), "vppcfg", "_version.py")
with open(version_file) as f:
    exec(f.read())

setup(
    version=__version__,
    packages=["vppcfg", "vppcfg/config", "vppcfg/vpp"],
    package_data={"vppcfg": ["*.yaml"]},
)



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""
The functions in this file interact with the VPP API to retrieve certain
interface metadata. Its base class will never change state. See the
derived classes VPPApiDumper() and VPPApiApplier()
"""

import os
import logging
import time
from vpp_papi import VPPApiClient, VPPApiJSONFiles, MACAddress


class VPPApi:
    """The VPPApi class is a base class that abstracts the vpp_papi."""

    def __init__(
        self,
        vpp_api_socket="/run/vpp/api.sock",
        vpp_json_dir=None,
        clientname="vppcfg",
    ):
        self.logger = logging.getLogger("vppcfg.vppapi")
        self.logger.addHandler(logging.NullHandler())

        self.vpp_api_socket = vpp_api_socket
        self.vpp_json_dir = vpp_json_dir
        self.vpp_jsonfiles = []
        self.vpp_messages = {}
        self.connected = False
        self.clientname = clientname
        self.vpp = None
        self.cache_read = False
        self.cache_clear()
        self.lcp_enabled = False

        if self.vpp_json_dir is None:
            self.vpp_json_dir = VPPApiJSONFiles.find_api_dir([])
        elif not os.path.isdir(self.vpp_json_dir):
            self.logger.error(f"VPP API JSON directory not found: {self.vpp_json_dir}")

        # Construct a list of all the JSON API files
        self.vpp_jsonfiles = VPPApiJSONFiles.find_api_files(api_dir=self.vpp_json_dir)
        if not self.vpp_jsonfiles:
            self.logger.error("No JSON API files found")

        # Enumerate all VPPMessage signatures from the JSON API files, and give their
        # API namedtuple defaults so creating instances can set only those fields which
        # are relevant.
        for json_filename in self.vpp_jsonfiles:
            with open(json_filename, "r", encoding="utf-8") as file_handle:
                for name, msg in VPPApiJSONFiles.process_json_file(file_handle)[
                    0
                ].items():
                    msg.tuple.__new__.__defaults__ = (None,) * len(msg.tuple._fields)
                    self.vpp_messages[name] = msg

    def connect(self, retries=30):
        """Connect to the VPP Dataplane, if we're not already connected"""
        if self.connected:
            return True

        if not os.path.exists(self.vpp_api_socket):
            self.logger.error(f"VPP api socket file not found: {self.vpp_api_socket}")
            return False

        self.vpp = VPPApiClient(
            apifiles=self.vpp_jsonfiles, server_address=self.vpp_api_socket
        )
        self.logger.debug("Connecting to VPP")
        for i in range(retries):
            try:
                self.vpp.connect(self.clientname)
                self.connected = True
                break
            except ConnectionError as err:
                self.logger.warning(
                    f"Could not connect to VPP (attempt {i+1}/{retries}): {err}"
                )
                time.sleep(1)
                self.connected = False
        if not self.connected:
            self.logger.error(f"Could not connect to VPP (tried {retries} times)")
            return False

        # pylint: disable=no-member
        api_response = self.vpp.api.show_version()
        self.logger.info(f"VPP version is {api_response.version}")

        return True

    def disconnect(self):
        """Disconnect from the VPP dataplane, if we are still connected."""
        if not self.connected:
            return True
        self.vpp.disconnect()
        self.logger.debug("Disconnected from VPP")
        self.connected = False
        return True

    def cache_clear(self):
        """Remove the cached VPP configuration elements and return True"""
        self.cache_read = False
        self.cache = {
            "lcps": {},
            "interface_names": {},
            "interfaces": {},
            "interface_addresses": {},
            "interface_unnumbered": {},
            "interface_mpls": {},
            "interface_acls": {},
            "bondethernets": {},
            "bondethernet_members": {},
            "bridgedomains": {},
            "vxlan_tunnels": {},
            "l2xcs": {},
            "taps": {},
            "acls": {},
            "acl_tags": {},
            "interface_sflow": {},
            "sflow": {},
        }
        return True

    def cache_remove_lcp(self, lcpname):
        """Removes the LCP and TAP interface, identified by lcpname, from the VPP config cache"""
        for _idx, lcp in self.cache["lcps"].items():
            if lcp.host_if_name == lcpname:
                ifname = self.cache["interfaces"][lcp.host_sw_if_index].interface_name
                del self.cache["lcps"][lcp.phy_sw_if_index]
                return self.cache_remove_interface(ifname)
        self.logger.warning(
            f"Trying to remove an LCP which is not in the config: {lcpname}"
        )
        return False

    def cache_remove_bondethernet_member(self, ifname):
        """Removes the bonderthernet member interface, identified by name, from the VPP config cache"""

        iface = self.get_interface_by_name(ifname)
        if not iface:
            self.logger.warning(
                f"Trying to remove a bondethernet member interface which is not in the config: {ifname}"
            )
            return False

        for bond_idx, members in self.cache["bondethernet_members"].items():
            if iface.sw_if_index in members:
                self.cache["bondethernet_members"][bond_idx].remove(iface.sw_if_index)

        return True

    def cache_remove_l2xc(self, ifname):
        """Remvoes the l2xc from the VPP config cache"""

        iface = self.get_interface_by_name(ifname)
        if not iface:
            self.logger.warning(
                f"Trying to remove an L2XC which is not in the config: {ifname}"
            )
            return False

        self.cache["l2xcs"].pop(iface.sw_if_index, None)
        return True

    def cache_remove_vxlan_tunnel(self, ifname):
        """Removes a vxlan_tunnel from the VPP config cache"""

        iface = self.get_interface_by_name(ifname)
        if not iface:
            self.logger.warning(
                f"Trying to remove a VXLAN Tunnel which is not in the config: {ifname}"
            )
            return False

        self.cache["vxlan_tunnels"].pop(iface.sw_if_index, None)
        return True

    def cache_remove_interface(self, ifname):
        """Removes the interface, identified by name, from the VPP config cache"""

        iface = self.get_interface_by_name(ifname)
        if not iface:
            self.logger.warning(
                f"Trying to remove an interface which is not in the config: {ifname}"
            )
            return False

        del self.cache["interfaces"][iface.sw_if_index]
        if len(self.cache["interface_addresses"][iface.sw_if_index]) > 0:
            self.logger.warning(f"Not all addresses were removed on {ifname}")
        del self.cache["interface_addresses"][iface.sw_if_index]
        self.cache["interface_acls"].pop(iface.sw_if_index, None)
        del self.cache["interface_names"][ifname]

        ## Use my_dict.pop('key', None), as it allows 'key' to be absent
        if iface.sw_if_index in self.cache["bondethernet_members"]:
            if len(self.cache["bondethernet_members"][iface.sw_if_index]) != 0:
                self.logger.warning(
                    f"When removing BondEthernet {ifname}, its members are not empty: {self.cache['bondethernet_members'][iface.sw_if_index]}"
                )
            else:
                del self.cache["bondethernet_members"][iface.sw_if_index]
        self.cache["bondethernets"].pop(iface.sw_if_index, None)
        self.cache["taps"].pop(iface.sw_if_index, None)
        return True

    def mockconfig(self, yaml_config):
        """Mock a minimal configuration cache without talking to a running VPP Dataplane, by
        enumerating the 'interfaces' scope from yaml_config"""

        if not "interfaces" in yaml_config:
            self.logger.error("YAML config does not contain any interfaces")
            return False
        self.logger.debug(f"config: {yaml_config['interfaces']}")

        self.cache_clear()
        ## Add mock local0
        idx = 0
        self.cache["interfaces"][idx] = self.vpp_messages["sw_interface_details"].tuple(
            sw_if_index=idx,
            sup_sw_if_index=idx,
            l2_address=MACAddress("00:00:00:00:00:00"),
            flags=0,
            type=0,
            link_duplex=0,
            link_speed=0,
            sub_id=0,
            sub_number_of_tags=0,
            sub_outer_vlan_id=0,
            sub_inner_vlan_id=0,
            sub_if_flags=0,
            vtr_op=0,
            vtr_push_dot1q=0,
            vtr_tag1=0,
            vtr_tag2=0,
            outer_tag=0,
            link_mtu=0,
            mtu=[0, 0, 0, 0],
            interface_name="local0",
            interface_dev_type="local",
            tag="mock",
        )
        self.cache["interface_acls"][idx] = self.vpp_messages[
            "acl_interface_list_details"
        ].tuple(
            sw_if_index=idx,
            count=0,
            n_input=0,
            acls=[],
        )
        ## Add mock PHYs
        for ifname, iface in yaml_config["interfaces"].items():
            if not "device-type" in iface or iface["device-type"] not in ["dpdk"]:
                continue
            idx += 1
            self.cache["interfaces"][idx] = self.vpp_messages[
                "sw_interface_details"
            ].tuple(
                sw_if_index=idx,
                sup_sw_if_index=idx,
                l2_address=MACAddress("00:00:00:00:00:00"),
                flags=0,
                type=0,
                link_duplex=0,
                link_speed=0,
                sub_id=0,
                sub_number_of_tags=0,
                sub_outer_vlan_id=0,
                sub_inner_vlan_id=0,
                sub_if_flags=0,
                vtr_op=0,
                vtr_push_dot1q=0,
                vtr_tag1=0,
                vtr_tag2=0,
                outer_tag=0,
                link_mtu=64,
                mtu=[64, 0, 0, 0],
                interface_name=ifname,
                interface_dev_type=iface["device-type"],
                tag="mock",
            )
            self.cache["interface_acls"][idx] = self.vpp_messages[
                "acl_interface_list_details"
            ].tuple(
                sw_if_index=idx,
                count=0,
                n_input=0,
                acls=[],
            )

        ## Create interface_names and interface_address indexes
        for idx, iface in self.cache["interfaces"].items():
            self.cache["interface_names"][iface.interface_name] = idx
            self.cache["interface_addresses"][idx] = []

        self.logger.debug(f"cache(mock): {self.cache}")
        return True

    def readconfig(self):
        """Read the configuration out of a running VPP Dataplane and put it into a
        VPP config cache"""
        # pylint: disable=no-member
        if not self.connected and not self.connect():
            self.logger.error("Could not connect to VPP")
            return False

        self.cache_clear()

        self.lcp_enabled = False
        try:
            self.logger.debug("Retrieving LCPs")
            api_response = self.vpp.api.lcp_itf_pair_get()
            if isinstance(api_response, tuple) and api_response[0].retval == 0:
                for lcp in api_response[1]:
                    self.cache["lcps"][lcp.phy_sw_if_index] = lcp
                self.lcp_enabled = True
        except AttributeError as err:
            self.logger.warning(f"LinuxCP API not found - missing plugin: {err}")

        self.logger.debug("Retrieving interfaces")
        api_response = self.vpp.api.sw_interface_dump()
        for iface in api_response:
            self.cache["interfaces"][iface.sw_if_index] = iface
            self.cache["interface_names"][iface.interface_name] = iface.sw_if_index
            self.cache["interface_addresses"][iface.sw_if_index] = []
            self.logger.debug(f"Retrieving IPv4 addresses for {iface.interface_name}")
            ipr = self.vpp.api.ip_address_dump(
                sw_if_index=iface.sw_if_index, is_ipv6=False
            )
            for addr in ipr:
                self.cache["interface_addresses"][iface.sw_if_index].append(
                    str(addr.prefix)
                )
            self.logger.debug(f"Retrieving IPv6 addresses for {iface.interface_name}")
            ipr = self.vpp.api.ip_address_dump(
                sw_if_index=iface.sw_if_index, is_ipv6=True
            )
            for addr in ipr:
                self.cache["interface_addresses"][iface.sw_if_index].append(
                    str(addr.prefix)
                )

        try:  ## TODO(pim): Remove after 23.10 release
            self.logger.debug("Retrieving interface MPLS state")
            api_response = self.vpp.api.mpls_interface_dump()
            for iface in api_response:
                self.cache["interface_mpls"][iface.sw_if_index] = True
        except AttributeError:
            self.logger.warning(
                "MPLS state retrieval requires https://gerrit.fd.io/r/c/vpp/+/39022"
            )

        try:
            self.logger.debug("Retrieving ACLs")
            api_response = self.vpp.api.acl_dump(acl_index=0xFFFFFFFF)
            for acl in api_response:
                self.cache["acls"][acl.acl_index] = acl

            self.logger.debug("Retrieving interface ACLs")
            api_response = self.vpp.api.acl_interface_list_dump()
            for iface in api_response:
                self.cache["interface_acls"][iface.sw_if_index] = iface
        except AttributeError as err:
            self.logger.warning(f"ACL API not found - missing plugin: {err}")

        self.logger.debug("Retrieving interface Unnumbered state")
        api_response = self.vpp.api.ip_unnumbered_dump()
        for iface in api_response:
            self.cache["interface_unnumbered"][iface.sw_if_index] = iface.ip_sw_if_index

        self.logger.debug("Retrieving bondethernets")
        api_response = self.vpp.api.sw_bond_interface_dump()
        for iface in api_response:
            self.cache["bondethernets"][iface.sw_if_index] = iface
            self.cache["bondethernet_members"][iface.sw_if_index] = []
            for member in self.vpp.api.sw_member_interface_dump(
                sw_if_index=iface.sw_if_index
            ):
                self.cache["bondethernet_members"][iface.sw_if_index].append(
                    member.sw_if_index
                )

        self.logger.debug("Retrieving bridgedomains")
        api_response = self.vpp.api.bridge_domain_dump()
        for bridge in api_response:
            self.cache["bridgedomains"][bridge.bd_id] = bridge

        try:
            self.logger.debug("Retrieving vxlan_tunnels")
            api_response = self.vpp.api.vxlan_tunnel_v2_dump()
            for vxlan in api_response:
                self.cache["vxlan_tunnels"][vxlan.sw_if_index] = vxlan
        except AttributeError as err:
            self.logger.warning(f"VXLAN API not found - missing plugin: {err}")

        self.logger.debug("Retrieving L2 Cross Connects")
        api_response = self.vpp.api.l2_xconnect_dump()
        for l2xc in api_response:
            self.cache["l2xcs"][l2xc.rx_sw_if_index] = l2xc

        self.logger.debug("Retrieving TAPs")
        api_response = self.vpp.api.sw_interface_tap_v2_dump()
        for tap in api_response:
            self.cache["taps"][tap.sw_if_index] = tap

        try:
            self.logger.debug("Retrieving sFlow")

            api_response = self.vpp.api.sflow_sampling_rate_get()
            if api_response:
                self.cache["sflow"]["sampling-rate"] = api_response.sampling_N
            api_response = self.vpp.api.sflow_polling_interval_get()
            if api_response:
                self.cache["sflow"]["polling-interval"] = api_response.polling_S
            api_response = self.vpp.api.sflow_header_bytes_get()
            if api_response:
                self.cache["sflow"]["header-bytes"] = api_response.header_B

            api_response = self.vpp.api.sflow_interface_dump()
            for iface in api_response:
                self.cache["interface_sflow"][iface.hw_if_index] = True
        except AttributeError as err:
            self.logger.warning(f"sFlow API not found - missing plugin: {err}")

        self.logger.debug("Retrieving interface Unnumbered state")
        api_response = self.vpp.api.ip_unnumbered_dump()
        for iface in api_response:
            self.cache["interface_unnumbered"][iface.sw_if_index] = iface.ip_sw_if_index

        self.logger.debug("Retrieving bondethernets")
        api_response = self.vpp.api.sw_bond_interface_dump()

        self.cache_read = True
        return self.cache_read

    def phys_exist(self, ifname_list):
        """Return True if all interfaces in the `ifname_list` exist as physical interface names
        in VPP. Return False otherwise."""
        ret = True
        for ifname in ifname_list:
            if not ifname in self.cache["interface_names"]:
                self.logger.warning(f"Interface {ifname} does not exist in VPP")
                ret = False
        return ret

    def get_interface_by_name(self, name):
        """Return the VPP interface specified by name, or None if it cannot be found"""
        try:
            idx = self.cache["interface_names"][name]
            return self.cache["interfaces"][idx]
        except KeyError:
            pass
        return None

    def get_sub_interfaces(self):
        """Return all interfaces which have a sub-id and one or more tags"""
        subints = [
            self.cache["interfaces"][x].interface_name
            for x in self.cache["interfaces"]
            if self.cache["interfaces"][x].sub_id > 0
            and self.cache["interfaces"][x].sub_number_of_tags > 0
        ]
        return subints

    def get_qinx_interfaces(self):
        """Return all interfaces which have a sub-id and a non-zero inner vlan tag"""
        qinx_subints = [
            self.cache["interfaces"][x].interface_name
            for x in self.cache["interfaces"]
            if self.cache["interfaces"][x].sub_id > 0
            and self.cache["interfaces"][x].sub_inner_vlan_id > 0
        ]
        return qinx_subints

    def get_dot1x_interfaces(self):
        """Return all interfaces which have only an outer vlan tag (dot1q/dot1ad)"""
        dot1x_subints = [
            self.cache["interfaces"][x].interface_name
            for x in self.cache["interfaces"]
            if self.cache["interfaces"][x].sub_id > 0
            and self.cache["interfaces"][x].sub_inner_vlan_id == 0
        ]
        return dot1x_subints

    def get_loopbacks(self):
        """Return all interfaces of VPP type 'Loopback'"""
        loopbacks = [
            self.cache["interfaces"][x].interface_name
            for x in self.cache["interfaces"]
            if self.cache["interfaces"][x].interface_dev_type == "Loopback"
        ]
        return loopbacks

    def get_phys(self):
        """Return all interfaces for which the super interface has the same sw_if_index
        and aren't known to be virtual interfaces"""
        lcp_tap_sw_if_indexes = {lcp.host_sw_if_index for lcp in self.cache["lcps"].values()}
        phys = [
            self.cache["interfaces"][x].interface_name
            for x in self.cache["interfaces"]
            if self.cache["interfaces"][x].sw_if_index
            == self.cache["interfaces"][x].sup_sw_if_index
            and self.cache["interfaces"][x].interface_dev_type
            not in ["virtio", "BVI", "Loopback", "VXLAN", "local", "bond"]
            and self.cache["interfaces"][x].sw_if_index not in lcp_tap_sw_if_indexes
        ]
        return phys

    def get_bondethernets(self):
        """Return all bondethernet interfaces"""
        bonds = [
            self.cache["bondethernets"][x].interface_name
            for x in self.cache["bondethernets"]
        ]
        return bonds

    def get_vxlan_tunnels(self):
        """Return all vxlan_tunnel interfaces"""
        vxlan_tunnels = [
            self.cache["interfaces"][x].interface_name
            for x in self.cache["interfaces"]
            if self.cache["interfaces"][x].interface_dev_type in ["VXLAN"]
        ]
        return vxlan_tunnels

    def get_lcp_by_interface(self, sw_if_index):
        """Return the LCP config cache for the interface given by sw_if_index"""
        for _idx, lcp in self.cache["lcps"].items():
            if lcp.phy_sw_if_index == sw_if_index:
                return lcp
        return None

    def tap_is_lcp(self, tap_ifname):
        """Returns True if the given tap_ifname is a TAP interface belonging to an LCP,
        or False otherwise."""

        vpp_iface = self.get_interface_by_name(tap_ifname)
        if not vpp_iface:
            return False

        for _idx, lcp in self.cache["lcps"].items():
            if vpp_iface.sw_if_index == lcp.host_sw_if_index:
                return True
        return False

    def cli_inband(self, cmd):
        """Call the VPP inband CLI with the given command, and return any retun value or False if we
        could not connect."""

        if not self.connected and not self.connect():
            self.logger.error("Could not connect to VPP")
            return False

        ret = self.vpp.api.cli_inband(cmd=cmd)
        return ret



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""Unit tests for vxlan_tunnels"""
import unittest
import yaml
from . import vxlan_tunnel
from .unittestyaml import UnitTestYaml


class TestVXLANMethods(unittest.TestCase):
    def setUp(self):
        with UnitTestYaml("test_vxlan_tunnel.yaml") as f:
            self.cfg = yaml.load(f, Loader=yaml.FullLoader)

    def test_get_by_name(self):
        ifname, iface = vxlan_tunnel.get_by_name(self.cfg, "vxlan_tunnel0")
        self.assertIsNotNone(iface)
        self.assertEqual("vxlan_tunnel0", ifname)

        ifname, iface = vxlan_tunnel.get_by_name(self.cfg, "vxlan_tunnel-noexist")
        self.assertIsNone(ifname)
        self.assertIsNone(iface)

    def test_is_vxlan_tunnel(self):
        self.assertTrue(vxlan_tunnel.is_vxlan_tunnel(self.cfg, "vxlan_tunnel0"))
        self.assertFalse(vxlan_tunnel.is_vxlan_tunnel(self.cfg, "vxlan_tunnel-noexist"))
        self.assertFalse(vxlan_tunnel.is_vxlan_tunnel(self.cfg, "GigabitEthernet1/0/0"))

    def test_enumerators(self):
        ifs = vxlan_tunnel.get_vxlan_tunnels(self.cfg)
        self.assertEqual(len(ifs), 4)
        self.assertIn("vxlan_tunnel0", ifs)
        self.assertIn("vxlan_tunnel1", ifs)
        self.assertIn("vxlan_tunnel2", ifs)
        self.assertIn("vxlan_tunnel3", ifs)
        self.assertNotIn("vxlan_tunnel-noexist", ifs)

    def test_vni_unique(self):
        self.assertTrue(vxlan_tunnel.vni_unique(self.cfg, 100))
        self.assertFalse(vxlan_tunnel.vni_unique(self.cfg, 101))
        self.assertTrue(vxlan_tunnel.vni_unique(self.cfg, 102))



# Copyright (c) 2025 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""
Planner prune operations - handles removal of objects from VPP that don't exist in config.
"""

from vppcfg.config import loopback
from vppcfg.config import interface
from vppcfg.config import bondethernet
from vppcfg.config import bridgedomain
from vppcfg.config import vxlan_tunnel
from vppcfg.config import tap
from .planner_base import PlannerBase


class PlannerPruneOperations(PlannerBase):  # pylint: disable=too-few-public-methods
    """Mixin class providing prune operations for the Planner."""

    def prune(self):
        """Remove all objects from VPP that do not occur in the config. For an indepth explanation
        of how and why this particular pruning order is chosen, see README.md section on
        Planning."""
        ret = True
        if not self._prune_admin_state():
            self.logger.warning("Could not set interfaces down in VPP")
            ret = False
        if not self._prune_lcps():
            self.logger.warning("Could not prune LCPs from VPP")
            ret = False
        if not self._prune_bridgedomains():
            self.logger.warning("Could not prune BridgeDomains from VPP")
            ret = False
        if not self._prune_loopbacks():
            self.logger.warning("Could not prune Loopbacks from VPP")
            ret = False
        if not self._prune_l2xcs():
            self.logger.warning("Could not prune L2 Cross Connects from VPP")
            ret = False
        if not self._prune_sub_interfaces():
            self.logger.warning("Could not prune Sub Interfaces from VPP")
            ret = False
        if not self._prune_taps():
            self.logger.warning("Could not prune TAPs from VPP")
            ret = False
        if not self._prune_vxlan_tunnels():
            self.logger.warning("Could not prune VXLAN Tunnels from VPP")
            ret = False
        if not self._prune_bondethernets():
            self.logger.warning("Could not prune BondEthernets from VPP")
            ret = False
        if not self._prune_phys():
            self.logger.warning("Could not prune PHYs from VPP")
            ret = False
        return ret

    def _prune_unnumbered_usage(self, target_ifname):
        """Remove the unnumbered use of all VPP interfaces that are using the given 'target_ifname'."""
        target_iface = self.vpp.get_interface_by_name(target_ifname)

        for idx, target_idx in self.vpp.cache["interface_unnumbered"].items():
            if target_idx == target_iface.sw_if_index:
                unnumbered_ifname = self.vpp.cache["interfaces"][idx].interface_name
                cli = f"set interface unnumbered del {unnumbered_ifname}"
                self.cli["prune"].append(cli)
        return True

    def _prune_addresses(self, ifname, address_list):
        """Remove all addresses from interface ifname, except those in address_list,
        which may be an empty list, in which case all addresses are removed.
        """
        _iface = self.vpp.get_interface_by_name(ifname)
        if not _iface:
            self.logger.error(
                f"Trying to prune addresses from non-existent interface {ifname}"
            )
            return

        idx = _iface.sw_if_index
        removed_addresses = []
        for addr in self.vpp.cache["interface_addresses"][idx]:
            if not addr in address_list:
                cli = f"set interface ip address del {ifname} {addr}"
                self.cli["prune"].append(cli)
                removed_addresses.append(addr)
            else:
                self.logger.debug(f"Address OK: {ifname} {addr}")
        for addr in removed_addresses:
            self.vpp.cache["interface_addresses"][idx].remove(addr)

    def _prune_loopbacks(self):
        """Remove loopbacks from VPP, if they do not occur in the config."""
        removed_interfaces = []
        for numtags in [2, 1, 0]:
            for _idx, vpp_iface in self.vpp.cache["interfaces"].items():
                if vpp_iface.interface_dev_type != "Loopback":
                    continue
                if vpp_iface.sub_number_of_tags != numtags:
                    continue
                _config_ifname, config_iface = loopback.get_by_name(
                    self.cfg, vpp_iface.interface_name
                )
                if not config_iface:
                    self._prune_addresses(vpp_iface.interface_name, [])
                    self._prune_unnumbered_usage(vpp_iface.interface_name)
                    if numtags == 0:
                        cli = f"delete loopback interface intfc {vpp_iface.interface_name}"
                        self.cli["prune"].append(cli)
                        removed_interfaces.append(vpp_iface.interface_name)
                    else:
                        cli = f"delete sub {vpp_iface.interface_name}"
                        self.cli["prune"].append(cli)
                        removed_interfaces.append(vpp_iface.interface_name)
                    continue
                self.logger.debug(f"Loopback OK: {vpp_iface.interface_name}")
                addresses = []
                if "addresses" in config_iface:
                    addresses = config_iface["addresses"]
                self._prune_addresses(vpp_iface.interface_name, addresses)

        for ifname in removed_interfaces:
            self.vpp.cache_remove_interface(ifname)

        return True

    def _prune_bridgedomains(self):
        """Remove bridge-domains from VPP, if they do not occur in the config. If any interfaces are
        found in to-be removed bridge-domains, they are returned to L3 mode, and tag-rewrites removed.
        """
        for idx, bridge in self.vpp.cache["bridgedomains"].items():
            bridgename = f"bd{int(idx)}"
            _config_ifname, config_iface = bridgedomain.get_by_name(
                self.cfg, bridgename
            )
            if not config_iface:
                for member in bridge.sw_if_details:
                    if member.sw_if_index == bridge.bvi_sw_if_index:
                        continue
                    member_iface = self.vpp.cache["interfaces"][member.sw_if_index]
                    member_ifname = member_iface.interface_name
                    if member_iface.sub_id > 0:
                        cli = f"set interface l2 tag-rewrite {member_ifname} disable"
                        self.cli["prune"].append(cli)
                    cli = f"set interface l3 {member_ifname}"
                    self.cli["prune"].append(cli)
                if bridge.bvi_sw_if_index in self.vpp.cache["interfaces"]:
                    bviname = self.vpp.cache["interfaces"][
                        bridge.bvi_sw_if_index
                    ].interface_name
                    cli = f"set interface l3 {bviname}"
                    self.cli["prune"].append(cli)
                cli = f"create bridge-domain {int(idx)} del"
                self.cli["prune"].append(cli)
            else:
                self.logger.debug(f"BridgeDomain OK: {bridgename}")
                for member in bridge.sw_if_details:
                    member_ifname = self.vpp.cache["interfaces"][
                        member.sw_if_index
                    ].interface_name
                    if (
                        "members" in config_iface
                        and member_ifname in config_iface["members"]
                    ):
                        if interface.is_sub(self.cfg, member_ifname):
                            cli = (
                                f"set interface l2 tag-rewrite {member_ifname} disable"
                            )
                            self.cli["prune"].append(cli)
                        cli = f"set interface l3 {member_ifname}"
                        self.cli["prune"].append(cli)
                if (
                    "bvi" in config_iface
                    and bridge.bvi_sw_if_index in self.vpp.cache["interfaces"]
                ):
                    bviname = self.vpp.cache["interfaces"][
                        bridge.bvi_sw_if_index
                    ].interface_name
                    if bviname != config_iface["bvi"]:
                        cli = f"set interface l3 {bviname}"
                        self.cli["prune"].append(cli)

        return True

    def _prune_l2xcs(self):
        """Remove all L2XC source interfaces from VPP, if they do not occur in the config. If they occur,
        but are crossconnected to a different interface name, also remove them. Interfaces are put
        back into L3 mode, and their tag-rewrites removed."""
        removed_l2xcs = []
        for _idx, l2xc in self.vpp.cache["l2xcs"].items():
            vpp_rx_ifname = self.vpp.cache["interfaces"][
                l2xc.rx_sw_if_index
            ].interface_name
            config_rx_ifname, config_rx_iface = interface.get_by_name(
                self.cfg, vpp_rx_ifname
            )
            if not config_rx_ifname:
                if self.vpp.cache["interfaces"][l2xc.rx_sw_if_index].sub_id > 0:
                    cli = f"set interface l2 tag-rewrite {vpp_rx_ifname} disable"
                    self.cli["prune"].append(cli)
                cli = f"set interface l3 {vpp_rx_ifname}"
                self.cli["prune"].append(cli)
                removed_l2xcs.append(vpp_rx_ifname)
                continue

            if not interface.is_l2xc_interface(self.cfg, config_rx_ifname):
                if interface.is_sub(self.cfg, config_rx_ifname):
                    cli = f"set interface l2 tag-rewrite {vpp_rx_ifname} disable"
                    self.cli["prune"].append(cli)
                cli = f"set interface l3 {vpp_rx_ifname}"
                self.cli["prune"].append(cli)
                removed_l2xcs.append(vpp_rx_ifname)
                continue
            vpp_tx_ifname = self.vpp.cache["interfaces"][
                l2xc.tx_sw_if_index
            ].interface_name
            if vpp_tx_ifname != config_rx_iface["l2xc"]:
                if interface.is_sub(self.cfg, config_rx_ifname):
                    cli = f"set interface l2 tag-rewrite {vpp_rx_ifname} disable"
                    self.cli["prune"].append(cli)
                cli = f"set interface l3 {vpp_rx_ifname}"
                self.cli["prune"].append(cli)
                removed_l2xcs.append(vpp_rx_ifname)
                continue
            self.logger.debug(f"L2XC OK: {vpp_rx_ifname} -> {vpp_tx_ifname}")
        for l2xc in removed_l2xcs:
            self.vpp.cache_remove_l2xc(l2xc)
        return True

    def _vxlan_tunnel_has_diff(self, ifname):
        """Returns True if the given ifname (vxlan_tunnel0) has different attributes between VPP
        and the given configuration, or if either does not exist.

        Returns False if they are identical."""

        vpp_iface = self.vpp.get_interface_by_name(ifname)
        if (
            not vpp_iface
            or vpp_iface.sw_if_index not in self.vpp.cache["vxlan_tunnels"]
        ):
            return True
        vpp_vxlan = self.vpp.cache["vxlan_tunnels"][vpp_iface.sw_if_index]

        _config_ifname, config_iface = vxlan_tunnel.get_by_name(self.cfg, ifname)
        if not config_iface:
            return True

        if config_iface["local"] != str(vpp_vxlan.src_address):
            return True
        if config_iface["remote"] != str(vpp_vxlan.dst_address):
            return True
        if config_iface["vni"] != vpp_vxlan.vni:
            return True
        return False

    def _tap_has_diff(self, ifname):  # pylint: disable=too-many-return-statements
        """Returns True if the given ifname (tap0) has different attributes between VPP
        and the given configuration, or if either does not exist.

        Returns False if the TAP is a Linux Control Plane LIP.
        Returns False if they are identical."""

        vpp_iface = self.vpp.get_interface_by_name(ifname)
        vpp_tap = self.vpp.cache["taps"][vpp_iface.sw_if_index]
        if not vpp_iface:
            return True

        _config_ifname, config_iface = tap.get_by_name(self.cfg, ifname)
        if not config_iface:
            return True

        if self.vpp.tap_is_lcp(ifname):
            return False

        if (
            "name" in config_iface["host"]
            and config_iface["host"]["name"] != vpp_tap.host_if_name
        ):
            return True
        if (
            "mtu" in config_iface["host"]
            and config_iface["host"]["mtu"] != vpp_tap.host_mtu_size
        ):
            return True
        if "mac" in config_iface["host"] and config_iface["host"]["mac"] != str(
            vpp_tap.host_mac_addr
        ):
            return True
        if (
            "bridge" in config_iface["host"]
            and config_iface["host"]["bridge"] != vpp_tap.host_bridge
        ):
            return True
        if (
            "namespace" in config_iface["host"]
            and config_iface["host"]["namespace"] != vpp_tap.host_namespace
        ):
            return True

        return False

    def _bond_has_diff(self, ifname):
        """Returns True if the given ifname (BondEthernet0) have different attributes,
        or if either does not exist.

        Returns False if they are identical.
        """

        vpp_iface = self.vpp.get_interface_by_name(ifname)
        if (
            not vpp_iface
            or not vpp_iface.sw_if_index in self.vpp.cache["bondethernets"]
        ):
            return True

        config_ifname, config_iface = bondethernet.get_by_name(self.cfg, ifname)
        if not config_iface:
            return True

        vpp_bond = self.vpp.cache["bondethernets"][vpp_iface.sw_if_index]
        mode = bondethernet.mode_to_int(bondethernet.get_mode(self.cfg, config_ifname))
        if mode not in (-1, vpp_bond.mode):
            return True
        loadbalance = bondethernet.lb_to_int(
            bondethernet.get_lb(self.cfg, config_ifname)
        )
        if loadbalance not in (-1, vpp_bond.lb):
            return True

        return False

    def _prune_taps(self):
        """Remove all TAPs from VPP, if they are not in the config. As an exception,
        TAPs which are a part of Linux Control Plane, are left alone, to be handled
        by _prune_lcps() later."""
        removed_taps = []
        for _idx, vpp_tap in self.vpp.cache["taps"].items():
            vpp_iface = self.vpp.cache["interfaces"][vpp_tap.sw_if_index]
            vpp_ifname = vpp_iface.interface_name
            if self.vpp.tap_is_lcp(vpp_ifname):
                continue
            if self._tap_has_diff(vpp_ifname):
                removed_taps.append(vpp_ifname)
                continue

        for ifname in removed_taps:
            cli = f"delete tap {ifname}"
            self.cli["prune"].append(cli)
            self.vpp.cache_remove_interface(ifname)
        return True

    def _prune_bondethernets(self):
        """Remove all BondEthernets from VPP, if they are not in the config. If the bond has members,
        remove those from the bond before removing the bond."""
        removed_interfaces = []
        removed_bondethernet_members = []
        for idx, bond in self.vpp.cache["bondethernets"].items():
            vpp_ifname = bond.interface_name
            _config_ifname, config_iface = bondethernet.get_by_name(
                self.cfg, vpp_ifname
            )

            if self._bond_has_diff(vpp_ifname):
                self._prune_addresses(vpp_ifname, [])
                self._prune_unnumbered_usage(vpp_ifname)
                for member in self.vpp.cache["bondethernet_members"][idx]:
                    member_ifname = self.vpp.cache["interfaces"][member].interface_name
                    cli = f"bond del {member_ifname}"
                    self.cli["prune"].append(cli)
                    removed_bondethernet_members.append(member_ifname)
                cli = f"delete bond {vpp_ifname}"
                self.cli["prune"].append(cli)
                removed_interfaces.append(vpp_ifname)
                continue

            for member in self.vpp.cache["bondethernet_members"][idx]:
                member_ifname = self.vpp.cache["interfaces"][member].interface_name
                if (
                    "interfaces" in config_iface
                    and not member_ifname in config_iface["interfaces"]
                ):
                    cli = f"bond del {member_ifname}"
                    self.cli["prune"].append(cli)
                    removed_bondethernet_members.append(member_ifname)
            addresses = []
            if "addresses" in config_iface:
                addresses = config_iface["addresses"]
            self._prune_addresses(vpp_ifname, addresses)
            self.logger.debug(f"BondEthernet OK: {vpp_ifname}")

        for ifname in removed_bondethernet_members:
            self.vpp.cache_remove_bondethernet_member(ifname)

        for ifname in removed_interfaces:
            self.vpp.cache_remove_interface(ifname)

        return True

    def _prune_vxlan_tunnels(self):
        """Remove all VXLAN Tunnels from VPP, if they are not in the config. If they are in the config
        but with differing attributes, remove them also."""
        removed_interfaces = []
        for idx, vpp_vxlan in self.vpp.cache["vxlan_tunnels"].items():
            vpp_ifname = self.vpp.cache["interfaces"][idx].interface_name
            config_ifname, config_iface = vxlan_tunnel.get_by_name(self.cfg, vpp_ifname)
            if not config_iface or self._vxlan_tunnel_has_diff(config_ifname):
                self._prune_addresses(vpp_ifname, [])
                cli = (
                    f"create vxlan tunnel instance {vpp_vxlan.instance} "
                    f"src {vpp_vxlan.src_address} dst {vpp_vxlan.dst_address} vni {vpp_vxlan.vni} del"
                )
                self.cli["prune"].append(cli)
                removed_interfaces.append(vpp_ifname)
                continue
            config_ifname, config_iface = interface.get_by_name(self.cfg, vpp_ifname)
            if config_iface:
                addresses = []
                if "addresses" in config_iface:
                    addresses = config_iface["addresses"]
                self._prune_addresses(vpp_ifname, addresses)
            self.logger.debug(f"VXLAN Tunnel OK: {vpp_ifname}")

        for ifname in removed_interfaces:
            self.vpp.cache_remove_vxlan_tunnel(ifname)
            self.vpp.cache_remove_interface(ifname)

        return True

    def _prune_sub_interfaces(self):
        """Remove interfaces from VPP if they are not in the config, if their encapsulation is different,
        or if the BondEthernet they reside on is different.
        Start with inner-most (QinQ/QinAD), then Dot1Q/Dot1AD."""
        removed_interfaces = []
        for numtags in [2, 1]:
            for vpp_ifname in self.vpp.get_sub_interfaces():
                vpp_iface = self.vpp.get_interface_by_name(vpp_ifname)
                if not vpp_iface or vpp_iface.sub_number_of_tags != numtags:
                    continue

                if self.vpp.tap_is_lcp(vpp_ifname):
                    continue

                prune = False
                _config_ifname, config_iface = interface.get_by_name(
                    self.cfg, vpp_ifname
                )
                if not config_iface:
                    prune = True
                elif (
                    vpp_iface.interface_dev_type == "bond"
                    and vpp_iface.sub_number_of_tags > 0
                ):
                    (
                        config_parent_ifname,
                        _config_parent_iface,
                    ) = interface.get_parent_by_name(self.cfg, vpp_ifname)
                    if self._bond_has_diff(config_parent_ifname):
                        prune = True

                config_encap = interface.get_encapsulation(self.cfg, vpp_ifname)
                vpp_encap = self._get_encapsulation(vpp_iface)
                if config_encap != vpp_encap:
                    prune = True

                if prune:
                    self._prune_addresses(vpp_ifname, [])
                    self._prune_unnumbered_usage(vpp_ifname)
                    cli = f"delete sub {vpp_ifname}"
                    self.cli["prune"].append(cli)
                    removed_interfaces.append(vpp_ifname)
                    continue

                addresses = []
                if "addresses" in config_iface:
                    addresses = config_iface["addresses"]
                self._prune_addresses(vpp_ifname, addresses)
                self.logger.debug(f"Sub Interface OK: {vpp_ifname}")

        for ifname in removed_interfaces:
            self.vpp.cache_remove_interface(ifname)

        return True

    def _prune_phys(self):
        """Set default MTU and remove IPs for PHYs that are not in the config."""
        for vpp_ifname in self.vpp.get_phys():
            vpp_iface = self.vpp.get_interface_by_name(vpp_ifname)
            if not vpp_iface:
                continue

            _config_ifname, config_iface = interface.get_by_name(self.cfg, vpp_ifname)
            if not config_iface:
                ## Interfaces were sent DOWN in the _prune_admin_state() step previously
                self._prune_addresses(vpp_ifname, [])
                if vpp_iface.link_mtu != 9000:
                    cli = f"set interface mtu 9000 {vpp_ifname}"
                    self.cli["prune"].append(cli)
                continue
            addresses = []
            if "addresses" in config_iface:
                addresses = config_iface["addresses"]
            self._prune_addresses(vpp_ifname, addresses)
            self.logger.debug(f"Interface OK: {vpp_ifname}")
        return True

    def _parent_iface_by_encap(self, sup_sw_if_index, outer, dot1ad=True):
        """Returns the sw_if_index of an interface on a given super_sw_if_index with given dot1q/dot1ad outer and inner-dot1q=0,
        in other words the intermediary Dot1Q/Dot1AD belonging to a QinX interface. If the interface doesn't exist, None is
        returned."""
        for idx, iface in self.vpp.cache["interfaces"].items():
            if iface.sup_sw_if_index != sup_sw_if_index:
                continue
            if iface.sub_inner_vlan_id > 0:
                continue
            if dot1ad and (iface.sub_if_flags & 8) and iface.sub_outer_vlan_id == outer:
                self.logger.debug(f"match: {iface.interface_name} (dot1ad)")
                return idx
            if (
                not dot1ad
                and not (iface.sub_if_flags & 8)
                and iface.sub_outer_vlan_id == outer
            ):
                self.logger.debug(f"match: {iface.interface_name} (dot1q)")
                return idx
        return None

    def _get_encapsulation(self, iface):
        """Return a dictionary-based encapsulation of the sub-interface, which helps comparing them to the same object
        returned by config.interface.get_encapsulation()."""
        if iface.sub_if_flags & 8:
            dot1ad = iface.sub_outer_vlan_id
            dot1q = 0
        else:
            dot1q = iface.sub_outer_vlan_id
            dot1ad = 0
        inner_dot1q = iface.sub_inner_vlan_id
        exact_match = iface.sub_if_flags & 16
        return {
            "dot1q": int(dot1q),
            "dot1ad": int(dot1ad),
            "inner-dot1q": int(inner_dot1q),
            "exact-match": bool(exact_match),
        }

    def _prune_lcps(self):
        """Remove LCPs which are not in the configuration, starting with QinQ/QinAD interfaces, then Dot1Q/Dot1AD,
        and finally PHYs/BondEthernets/Tunnels/Loopbacks. For QinX, special care is taken to ensure that
        their intermediary interface exists, and has the correct encalsulation. If the intermediary interface
        changed, the QinX LCP is removed. The same is true for Dot1Q/Dot1AD interfaces: if their encapsulation
        has changed, we will have to re-create the underlying sub-interface, so the LCP has to be removed.

        Order is important: destroying an LCP of a PHY will invalidate its Dot1Q/Dot1AD as well as their
        downstream children in Linux.
        """
        lcps = self.vpp.cache["lcps"]

        removed_lcps = []
        for numtags in [2, 1, 0]:
            for _idx, lcp_iface in lcps.items():
                vpp_iface = self.vpp.cache["interfaces"][lcp_iface.phy_sw_if_index]
                if vpp_iface.sub_number_of_tags != numtags:
                    continue
                if vpp_iface.interface_dev_type == "Loopback":
                    config_ifname, config_iface = loopback.get_by_lcp_name(
                        self.cfg, lcp_iface.host_if_name
                    )
                else:
                    config_ifname, config_iface = interface.get_by_lcp_name(
                        self.cfg, lcp_iface.host_if_name
                    )
                if not config_iface:
                    ## Interface doesn't exist in the config
                    removed_lcps.append(lcp_iface)
                    continue
                if not "lcp" in config_iface:
                    ## Interface doesn't have an LCP
                    removed_lcps.append(lcp_iface)
                    continue
                if vpp_iface.sub_number_of_tags == 2:
                    vpp_parent_idx = self._parent_iface_by_encap(
                        vpp_iface.sup_sw_if_index,
                        vpp_iface.sub_outer_vlan_id,
                        vpp_iface.sub_if_flags & 8,
                    )
                    vpp_parent_iface = self.vpp.cache["interfaces"][vpp_parent_idx]
                    parent_lcp = lcps[vpp_parent_iface.sw_if_index]
                    (
                        config_parent_ifname,
                        config_parent_iface,
                    ) = interface.get_by_lcp_name(self.cfg, parent_lcp.host_if_name)
                    if not config_parent_iface:
                        ## QinX's parent doesn't exist in the config
                        removed_lcps.append(lcp_iface)
                        continue
                    if not "lcp" in config_parent_iface:
                        ## QinX's parent doesn't have an LCP
                        removed_lcps.append(lcp_iface)
                        continue
                    if parent_lcp.host_if_name != config_parent_iface["lcp"]:
                        ## QinX's parent LCP name mismatch
                        removed_lcps.append(lcp_iface)
                        continue
                    config_parent_encap = interface.get_encapsulation(
                        self.cfg, config_parent_ifname
                    )
                    vpp_parent_encap = self._get_encapsulation(vpp_parent_iface)
                    if config_parent_encap != vpp_parent_encap:
                        ## QinX's parent encapsulation mismatch
                        removed_lcps.append(lcp_iface)
                        continue

                if vpp_iface.sub_number_of_tags > 0:
                    config_encap = interface.get_encapsulation(self.cfg, config_ifname)
                    vpp_encap = self._get_encapsulation(vpp_iface)
                    if config_encap != vpp_encap:
                        ## Encapsulation mismatch
                        removed_lcps.append(lcp_iface)
                        continue

                if vpp_iface.interface_dev_type == "Loopback":
                    ## Loopbacks will not have a PHY to check.
                    continue
                if vpp_iface.interface_dev_type == "bond":
                    bond_iface = self.vpp.cache["interfaces"][vpp_iface.sup_sw_if_index]
                    if self._bond_has_diff(bond_iface.interface_name):
                        ## If BondEthernet changed, it has to be re-created, so all LCPs must be removed.
                        removed_lcps.append(lcp_iface)
                        continue

                phy_lcp = lcps[vpp_iface.sup_sw_if_index]
                _config_phy_ifname, config_phy_iface = interface.get_by_lcp_name(
                    self.cfg, phy_lcp.host_if_name
                )
                if not config_phy_iface:
                    ## Phy doesn't exist in the config
                    removed_lcps.append(lcp_iface)
                    continue
                if not "lcp" in config_phy_iface:
                    ## Phy doesn't have an LCP
                    removed_lcps.append(lcp_iface)
                    continue
                if phy_lcp.host_if_name != config_phy_iface["lcp"]:
                    ## Phy LCP name mismatch
                    removed_lcps.append(lcp_iface)
                    continue

                self.logger.debug(
                    f"LCP OK: {lcp_iface.host_if_name} -> (vpp={vpp_iface.interface_name}, config={config_ifname})"
                )

        for lcp_iface in removed_lcps:
            vpp_ifname = self.vpp.cache["interfaces"][
                lcp_iface.phy_sw_if_index
            ].interface_name
            cli = f"lcp delete {vpp_ifname}"
            self.cli["prune"].append(cli)
            self.vpp.cache_remove_lcp(lcp_iface.host_if_name)
        return True

    def _prune_admin_state(self):
        """Set admin-state down for all interfaces that are not in the config."""
        for ifname in (
            self.vpp.get_qinx_interfaces()
            + self.vpp.get_dot1x_interfaces()
            + self.vpp.get_bondethernets()
            + self.vpp.get_phys()
            + self.vpp.get_vxlan_tunnels()
            + self.vpp.get_loopbacks()
        ):
            if not ifname in interface.get_interfaces(
                self.cfg
            ) + loopback.get_loopbacks(self.cfg):
                vpp_iface = self.vpp.get_interface_by_name(ifname)
                if not vpp_iface:
                    continue

                if self.vpp.tap_is_lcp(ifname):
                    continue

                if vpp_iface.flags & 1:  # IF_STATUS_API_FLAG_ADMIN_UP
                    cli = f"set interface state {ifname} down"
                    self.cli["prune"].append(cli)

        return True



#
# Copyright (c) 2022 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A vppcfg configuration module that handles bridgedomains"""
import logging
from . import interface
from . import loopback


def get_bridgedomains(yaml):
    """Return a list of all bridgedomains."""
    ret = []
    if not "bridgedomains" in yaml:
        return ret
    for ifname, _iface in yaml["bridgedomains"].items():
        ret.append(ifname)
    return ret


def get_by_name(yaml, ifname):
    """Return the BridgeDomain by name (bd*), if it exists. Return None,None otherwise."""
    try:
        if ifname in yaml["bridgedomains"]:
            return ifname, yaml["bridgedomains"][ifname]
    except KeyError:
        pass
    return None, None


def is_bridgedomain(yaml, ifname):
    """Returns True if the name (bd*) is an existing bridgedomain."""
    ifname, iface = get_by_name(yaml, ifname)
    return iface is not None


def get_bridge_interfaces(yaml):
    """Returns a list of all interfaces that are bridgedomain members"""

    ret = []
    if not "bridgedomains" in yaml:
        return ret

    for _ifname, iface in yaml["bridgedomains"].items():
        if "interfaces" in iface:
            ret.extend(iface["interfaces"])

    return ret


def is_bridge_interface_unique(yaml, ifname):
    """Returns True if this interface is referenced in bridgedomains zero or one times"""

    ifs = get_bridge_interfaces(yaml)
    return ifs.count(ifname) < 2


def is_bridge_interface(yaml, ifname):
    """Returns True if this interface is a member of a BridgeDomain"""

    return ifname in get_bridge_interfaces(yaml)


def bvi_unique(yaml, bviname):
    """Returns True if the BVI identified by bviname is unique among all BridgeDomains."""
    if not "bridgedomains" in yaml:
        return True
    ncount = 0
    for _ifname, iface in yaml["bridgedomains"].items():
        if "bvi" in iface and iface["bvi"] == bviname:
            ncount += 1
    return ncount < 2


def get_settings(yaml, ifname):
    """Return a dictionary of 'settings' including their VPP defaults, for the
    bridgedomain identified by 'ifname' (bd10)"""
    ifname, iface = get_by_name(yaml, ifname)
    if not iface:
        return None

    settings = {
        "learn": True,
        "unicast-flood": True,
        "unknown-unicast-flood": True,
        "unicast-forward": True,
        "arp-termination": False,
        "arp-unicast-forward": False,
        "mac-age-minutes": 0,  ## 0 means disabled
    }
    if "settings" in iface:
        if "learn" in iface["settings"]:
            settings["learn"] = iface["settings"]["learn"]
        if "unicast-flood" in iface["settings"]:
            settings["unicast-flood"] = iface["settings"]["unicast-flood"]
        if "unknown-unicast-flood" in iface["settings"]:
            settings["unknown-unicast-flood"] = iface["settings"][
                "unknown-unicast-flood"
            ]
        if "unicast-forward" in iface["settings"]:
            settings["unicast-forward"] = iface["settings"]["unicast-forward"]
        if "arp-termination" in iface["settings"]:
            settings["arp-termination"] = iface["settings"]["arp-termination"]
        if "arp-unicast-forward" in iface["settings"]:
            settings["arp-unicast-forward"] = iface["settings"]["arp-unicast-forward"]
        if "mac-age-minutes" in iface["settings"]:
            settings["mac-age-minutes"] = int(iface["settings"]["mac-age-minutes"])
    return settings


def validate_bridgedomains(yaml):
    """Validate the semantics of all YAML 'bridgedomains' entries"""
    result = True
    msgs = []
    logger = logging.getLogger("vppcfg.config")
    logger.addHandler(logging.NullHandler())

    if not "bridgedomains" in yaml:
        return result, msgs

    for ifname, iface in yaml["bridgedomains"].items():
        logger.debug(f"bridgedomain {iface}")
        bd_mtu = 1500
        if "mtu" in iface:
            bd_mtu = iface["mtu"]
        instance = int(ifname[2:])
        if instance == 0:
            msgs.append(f"bridgedomain {ifname} is reserved")
            result = False
        elif instance > 65535:
            msgs.append(
                f"bridgedomain {ifname} has instance {int(instance)} which is too large"
            )
            result = False

        if "bvi" in iface:
            bvi_ifname, bvi_iface = loopback.get_by_name(yaml, iface["bvi"])
            if not bvi_unique(yaml, bvi_ifname):
                msgs.append(f"bridgedomain {ifname} BVI {bvi_ifname} is not unique")
                result = False
            if not bvi_iface:
                msgs.append(f"bridgedomain {ifname} BVI {bvi_ifname} does not exist")
                result = False
                continue

            bvi_mtu = 1500
            if "mtu" in bvi_iface:
                bvi_mtu = bvi_iface["mtu"]
            if bvi_mtu != bd_mtu:
                msgs.append(
                    f"bridgedomain {ifname} BVI {bvi_ifname} has MTU {int(bvi_mtu)}, while bridge has {int(bd_mtu)}"
                )
                result = False

        if "interfaces" in iface:
            for member in iface["interfaces"]:
                if (None, None) == interface.get_by_name(yaml, member):
                    msgs.append(f"bridgedomain {ifname} member {member} does not exist")
                    result = False
                    continue

                if not is_bridge_interface_unique(yaml, member):
                    msgs.append(f"bridgedomain {ifname} member {member} is not unique")
                    result = False
                if interface.has_lcp(yaml, member):
                    msgs.append(f"bridgedomain {ifname} member {member} has an LCP")
                    result = False
                if interface.has_address(yaml, member):
                    msgs.append(f"bridgedomain {ifname} member {member} has an address")
                    result = False
                member_mtu = interface.get_mtu(yaml, member)
                if member_mtu != bd_mtu:
                    msgs.append(
                        f"bridgedomain {ifname} member {member} has MTU {int(member_mtu)}, while bridge has {int(bd_mtu)}"
                    )
                    result = False

    return result, msgs



# Copyright (c) 2025 Pim van Pelt
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -*- coding: utf-8 -*-
"""
The functions in this file interact with the VPP API to retrieve certain
metadata, and plan configuration changes towards a given YAML target configuration.
"""
import sys
import logging
from vppcfg.config import interface
from vppcfg.config import lcp
from .vppapi import VPPApi
from .planner_prune import PlannerPruneOperations
from .planner_create import PlannerCreateOperations
from .planner_sync import PlannerSyncOperations


class Planner(PlannerPruneOperations, PlannerCreateOperations, PlannerSyncOperations):
    """The Planner class first reads the running configuration of a VPP Dataplane,
    and based on an intended target YAML configuration file, plans a path to make the
    dataplane safely reflect the target config. It first prunes (removes) objects that
    are not meant to be in the dataplane, or are in the dataplane but are not of the
    correct create-time attributes; then it creates objects that are in the configuration
    but not yet in the dataplane; and finally it syncs the configuration attributes of
    objects that can be changed at runtime."""

    def __init__(
        self,
        cfg,
        vpp_api_socket="/run/vpp/api.sock",
        vpp_json_dir=None,
    ):
        super().__init__()
        self.logger = logging.getLogger("vppcfg.planner")
        self.logger.addHandler(logging.NullHandler())

        self.vpp = VPPApi(vpp_api_socket, vpp_json_dir)
        self.cfg = cfg

        ## List of CLI calls emitted during the prune, create and sync phases.
        self.cli = {"prune": [], "create": [], "sync": []}

    def __del__(self):
        self.vpp.disconnect()

    def lcps_exist_with_lcp_enabled(self):
        """Returns False if there are LCPs defined in the configuration, but LinuxCP
        functionality is not enabled in VPP."""
        if not lcp.get_lcps(self.cfg):
            return True
        return self.vpp.lcp_enabled

    def phys_exist_in_vpp(self):
        """Return True if all PHYs in the config exist as physical interface names
        in VPP. Return False otherwise."""

        ret = True
        for ifname in interface.get_phys(self.cfg):
            if not ifname in self.vpp.cache["interface_names"]:
                self.logger.warning(f"Interface {ifname} does not exist in VPP")
                ret = False
        return ret

    def phys_exist_in_config(self):
        """Return True if all interfaces in VPP exist as physical interface names
        in the config. Return False otherwise."""

        ret = True
        for ifname in self.vpp.get_phys():
            if not ifname in interface.get_interfaces(self.cfg):
                self.logger.warning(f"Interface {ifname} does not exist in the config")
                ret = False
        return ret

    def write(self, outfile, emit_ok=False):
        """Emit the CLI contents to stdout (if outfile=='-') or a named file otherwise.
        If the 'emit_ok' flag is False, emit a warning at the top and bottom of the file.
        """
        # Assemble the intended output into a list
        output = []
        if not emit_ok:
            output.append(
                "comment { vppcfg: Planning failed, be careful with this output! }"
            )

        for phase in ["prune", "create", "sync"]:
            ncount = len(self.cli[phase])
            if ncount > 0:
                output.append(
                    f"comment {{ vppcfg {phase}: {ncount} CLI statement(s) follow }}"
                )
                output.extend(self.cli[phase])

        if not emit_ok:
            output.append(
                "comment { vppcfg: Planning failed, be careful with this output! }"
            )

        # Emit the output list to stdout or a file
        if outfile and outfile == "-":
            file = sys.stdout
            outfile = "(stdout)"
            if len(output) > 0:
                print("\n".join(output), file=file)
        else:
            with open(outfile, "w", encoding="utf-8") as file:
                if len(output) > 0:
                    print("\n".join(output), file=file)

        self.logger.info(f"Wrote {len(output)} lines to {outfile}")

    def get_commands(self):
        """Returns the CLI commands as a dictionary."""
        return self.cli
